﻿/*
  Synthesizer

  簡易的なポリフォニックシンセサイザー。
  サンプリングデータも一緒に鳴らせる。

  ・DAC端子にアンプとスピーカーを適宜接続
  ・元ネタはAVR用のシンセ
  ・PCのキーボードを鍵盤に見立てて演奏
  ・256ステップのウェーブテーブルとEGによる音色(VCOやLFO類はない)
  ・PCM RAWファイルの再生(ベロシティーと簡易周波数UP/DOWNコンバート)
  ・周期的なタイマ割り込みでミキシングとDACの出力を更新、DMA化して余力
    を残せば周波数アップやフィルタの追加、ステレオ化も夢ではない
  ・バイナリファイルをソースから直埋め込みしてみた
  ・テーブルと無圧縮なRAWファイルだけでメモリを圧迫してしまうので、RAW
    ファイルぐらいは外部メモリから取り込めたら良いな
 */

#include  <us.h>
#include  "e097b_def.h"

// USART_TYPEマクロで使用するUSARTと機能を選択
// CN3/CN4     1:ポーリング, 2:割り込み, 3:DMA
// CN1-6,CN1-7 4:ポーリング, 5:割り込み, 6:DMA
#ifndef USART_TYPE
  #define USART_TYPE 3
#endif
#include  "e097b_usart.h"

#include  <stdlib.h>
#include  <string.h>
#include  <math.h>

// マクロ・グローバル変数など
//-----------------------------
#define _MAX_NOTES      (15)    // 最大和音

// PCM
//-----------------------------
// RAWファイルの埋め込み (11.025kHz Signed 8bit RAW C4想定)
// バイナリファイルの埋め込み方詳細はこちら https://gist.github.com/mmozeiko/ed9655cf50341553d282
// wavはこちらから拝借  https://freewavesamples.com/
// raw変換はsox   http://sox.sourceforge.net/
INCBIN ("kick.raw", pcm_kick);
INCBIN ("sd.raw", pcm_sn);
INCBIN ("hh2.raw", pcm_hh);
INCBIN ("cc.raw", pcm_cc);
INCBIN ("tom.raw", pcm_tom);
INCBIN ("hit2.raw", pcm_hit);

// PCMデータの管理
typedef struct {
  int8_t  *raw; // PCMデータポインタ
  int32_t size; // PCMデータバイトサイズ
  int32_t ind;  // -1:停止, 0:スタート
  int16_t freq; // 周波数 (256:基準)
  uint8_t velo; // 0..127
} TPCMInfo;

TPCMInfo PCMInfo[] = {
  { (int8_t *)&pcm_kick, (int)&_size_pcm_kick, -1, 256, 127 },
  { (int8_t *)&pcm_sn,   (int)&_size_pcm_sn,   -1, 256, 127 },
  { (int8_t *)&pcm_hh,   (int)&_size_pcm_hh,   -1, 256, 127 },
  { (int8_t *)&pcm_cc,   (int)&_size_pcm_cc,   -1, 256, 127 },
  { (int8_t *)&pcm_tom,  (int)&_size_pcm_tom,  -1, 256, 127 },
  { (int8_t *)&pcm_hit,  (int)&_size_pcm_hit,  -1, 256, 127 },
};

// PCMデータの再生を励起
// 再生そのものはMRTとDACで行う
void PCM_play (uint8_t ind, uint8_t tone, uint8_t velo) {
  const uint16_t freq_table[12] = {8192,8679,9195,9742,10321,10935,11585,12274,13004,13777,14596,15464};  // オクターブ9
  if (ind < (sizeof (PCMInfo) / sizeof (PCMInfo[0])) && tone <= 127) {
    div_t o_t = div (tone, 12);
    PCMInfo[ind].ind = -1;
    // freqによって再生周波数を変更(間延びさせるとも言う)
    PCMInfo[ind].freq = freq_table[o_t.rem] >> (9 - o_t.quot);
    PCMInfo[ind].velo = velo;
    PCMInfo[ind].ind = 0;
  }
}

// LED
//-----------------------------
#define SCT_PWM_PIN_LED    0    // COUT0
#define SCT_PWM_LED        1    // Index of LED PWM (1..)

void LED_Init (void) {
  const TPin p = { _LED2, PIO_TYPE_MOVABLE, SWM_SCTOUT_0_O, PIO_MODE_DEFAULT }; // PIO0_24    SCT0
  PIO_Configure (&p, 1);

  Chip_SCTPWM_Init (LPC_SCT);
  Chip_SCTPWM_SetRate (LPC_SCT, 10000); // 10kHz
  Chip_SCTPWM_SetOutPin (LPC_SCT, SCT_PWM_LED, SCT_PWM_PIN_LED);
  Chip_SCTPWM_SetDutyCycle (LPC_SCT, SCT_PWM_LED, Chip_SCTPWM_GetTicksPerCycle (LPC_SCT) / 2);
  Chip_SCTPWM_Start (LPC_SCT);
  Chip_SCTPWM_SetDutyCycle (LPC_SCT, SCT_PWM_LED, Chip_SCTPWM_GetTicksPerCycle (LPC_SCT));  // 消灯
}

void LED_SetBright (int permil) {
  permil = MAX (MIN (permil, 1000), 0);
  Chip_SCT_SetMatchReload (LPC_SCT, (CHIP_SCT_MATCH_REG_T)SCT_PWM_LED, Chip_SCTPWM_GetTicksPerCycle (LPC_SCT) - (Chip_SCTPWM_GetTicksPerCycle (LPC_SCT) * permil) / 1000);
}

bool led_on = false;
void LED_Toggle (void) {
  bool On = !led_on;
  LED_SetBright (On ? 30 : 0);
  led_on = On;
}

void LED_Set (bool On) {
  LED_SetBright (On ? 30 : 0);
  led_on = On;
}

// Synthesizer
// from https://bitluni.net/
//  arduinoMIDISynthをほぼそのまま引用
//-----------------------------
//starage for ADSR envelope
int8_t adsr[4] = { 1, 127, 0, 0 };
//interpolated signed wave samples
int8_t wave[256] = { 0 };

#include "fixedPoint.h"

//this frequency table is for 44100Hz
// for (int i = 0; i < 128; i++) v = UINT32_MAX / 44100.0 * (440.0 * pow (2.0, (i - 69) / 12.0))
const uint32_t freqIncTab[128] = {
      796254,     843601,     893765,     946911,    1003217,    1062871,    1126073,    1193033,    1263974,    1339134,    1418763,    1503127,
     1592507,    1687203,    1787529,    1893821,    2006434,    2125742,    2252146,    2386065,    2527948,    2678268,    2837526,    3006254,
     3185015,    3374406,    3575058,    3787642,    4012867,    4251485,    4504291,    4772130,    5055896,    5356535,    5675051,    6012507,
     6370030,    6748811,    7150117,    7575285,    8025735,    8502970,    9008582,    9544261,   10111792,   10713070,   11350103,   12025015,
    12740059,   13497623,   14300233,   15150569,   16051469,   17005939,   18017165,   19088521,   20223584,   21426141,   22700205,   24050030,
    25480119,   26995246,   28600467,   30301139,   32102938,   34011878,   36034330,   38177043,   40447168,   42852281,   45400411,   48100060,
    50960238,   53990491,   57200933,   60602278,   64205876,   68023757,   72068660,   76354085,   80894335,   85704563,   90800821,   96200119,
   101920476,  107980983,  114401866,  121204555,  128411753,  136047513,  144137319,  152708170,  161788671,  171409126,  181601643,  192400238,
   203840952,  215961966,  228803732,  242409110,  256823506,  272095026,  288274639,  305416341,  323577341,  342818251,  363203285,  384800477,
   407681904,  431923931,  457607465,  484818220,  513647012,  544190053,  576549277,  610832681,  647154683,  685636503,  726406571,  769600953,
   815363807,  863847862,  915214929,  969636441, 1027294024, 1088380105, 1153098554, 1221665363
};

//dynamic range compression table
// for (int i = 0; i < 4096; i++) v = atan ((i - 2048) * 0.006) / M_PI_2 * 134 + 128;
const uint8_t compressTab[4096] = {
  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
  1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
  2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
  2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
  2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,3,3,3,
  3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
  3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
  3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
  4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,5,5,5,
  5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
  5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,6,6,6,6,6,
  6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,
  6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
  7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
  8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,9,9,9,9,9,9,9,9,9,9,9,
  9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,10,10,10,10,10,10,10,10,10,10,10,
  10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,
  11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,
  12,12,12,12,12,12,12,12,12,12,12,12,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,14,14,14,14,14,14,14,14,14,14,14,14,
  14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,17,18,18,18,18,18,18,18,18,18,18,18,18,18,
  18,18,18,18,18,18,18,18,18,18,18,18,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,20,21,21,21,21,21,21,21,
  21,21,21,21,21,21,21,21,21,21,21,21,21,21,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,22,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,23,24,24,24,24,24,24,24,24,24,24,24,24,24,24,
  24,24,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,25,26,26,26,26,26,26,26,26,26,26,26,26,26,26,27,27,27,27,27,27,27,27,27,27,27,27,27,27,28,28,28,28,28,28,28,28,28,28,28,28,28,29,29,29,29,29,
  29,29,29,29,29,29,29,30,30,30,30,30,30,30,30,30,30,30,30,31,31,31,31,31,31,31,31,31,31,31,32,32,32,32,32,32,32,32,32,32,32,33,33,33,33,33,33,33,33,33,33,34,34,34,34,34,34,34,34,34,35,35,35,35,
  35,35,35,35,35,36,36,36,36,36,36,36,36,36,37,37,37,37,37,37,37,37,38,38,38,38,38,38,38,38,38,39,39,39,39,39,39,39,40,40,40,40,40,40,40,40,41,41,41,41,41,41,41,42,42,42,42,42,42,42,43,43,43,43,
  43,43,44,44,44,44,44,44,44,45,45,45,45,45,45,46,46,46,46,46,46,47,47,47,47,47,47,48,48,48,48,48,49,49,49,49,49,49,50,50,50,50,50,51,51,51,51,51,52,52,52,52,52,53,53,53,53,53,54,54,54,54,55,55,
  55,55,55,56,56,56,56,57,57,57,57,57,58,58,58,58,59,59,59,59,60,60,60,60,61,61,61,61,62,62,62,62,63,63,63,64,64,64,64,65,65,65,65,66,66,66,67,67,67,68,68,68,68,69,69,69,70,70,70,71,71,71,71,72,
  72,72,73,73,73,74,74,74,75,75,75,76,76,76,77,77,78,78,78,79,79,79,80,80,80,81,81,82,82,82,83,83,83,84,84,85,85,85,86,86,87,87,87,88,88,89,89,89,90,90,91,91,92,92,92,93,93,94,94,95,95,95,96,96,
  97,97,98,98,99,99,99,100,100,101,101,102,102,103,103,104,104,105,105,106,106,106,107,107,108,108,109,109,110,110,111,111,112,112,113,113,114,114,115,115,116,116,117,117,118,118,119,119,120,120,121,121,122,122,123,123,124,124,125,125,126,126,127,127,
  128,129,129,130,130,131,131,132,132,133,133,134,134,135,135,136,136,137,137,138,138,139,139,140,140,141,141,142,142,143,143,144,144,145,145,146,146,147,147,148,148,149,149,150,150,150,151,151,152,152,153,153,154,154,155,155,156,156,157,157,157,158,158,159,
  159,160,160,161,161,161,162,162,163,163,164,164,164,165,165,166,166,167,167,167,168,168,169,169,169,170,170,171,171,171,172,172,173,173,173,174,174,174,175,175,176,176,176,177,177,177,178,178,178,179,179,180,180,180,181,181,181,182,182,182,183,183,183,184,
  184,184,185,185,185,185,186,186,186,187,187,187,188,188,188,188,189,189,189,190,190,190,191,191,191,191,192,192,192,192,193,193,193,194,194,194,194,195,195,195,195,196,196,196,196,197,197,197,197,198,198,198,198,199,199,199,199,199,200,200,200,200,201,201,
  201,201,201,202,202,202,202,203,203,203,203,203,204,204,204,204,204,205,205,205,205,205,206,206,206,206,206,207,207,207,207,207,207,208,208,208,208,208,209,209,209,209,209,209,210,210,210,210,210,210,211,211,211,211,211,211,212,212,212,212,212,212,212,213,
  213,213,213,213,213,214,214,214,214,214,214,214,215,215,215,215,215,215,215,216,216,216,216,216,216,216,216,217,217,217,217,217,217,217,218,218,218,218,218,218,218,218,218,219,219,219,219,219,219,219,219,220,220,220,220,220,220,220,220,220,221,221,221,221,
  221,221,221,221,221,222,222,222,222,222,222,222,222,222,223,223,223,223,223,223,223,223,223,223,224,224,224,224,224,224,224,224,224,224,224,225,225,225,225,225,225,225,225,225,225,225,226,226,226,226,226,226,226,226,226,226,226,226,227,227,227,227,227,227,
  227,227,227,227,227,227,228,228,228,228,228,228,228,228,228,228,228,228,228,229,229,229,229,229,229,229,229,229,229,229,229,229,229,230,230,230,230,230,230,230,230,230,230,230,230,230,230,231,231,231,231,231,231,231,231,231,231,231,231,231,231,231,231,232,
  232,232,232,232,232,232,232,232,232,232,232,232,232,232,232,233,233,233,233,233,233,233,233,233,233,233,233,233,233,233,233,233,233,234,234,234,234,234,234,234,234,234,234,234,234,234,234,234,234,234,234,235,235,235,235,235,235,235,235,235,235,235,235,235,
  235,235,235,235,235,235,235,235,236,236,236,236,236,236,236,236,236,236,236,236,236,236,236,236,236,236,236,236,236,237,237,237,237,237,237,237,237,237,237,237,237,237,237,237,237,237,237,237,237,237,237,237,237,238,238,238,238,238,238,238,238,238,238,238,
  238,238,238,238,238,238,238,238,238,238,238,238,238,238,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,239,240,240,240,240,240,240,240,240,240,240,240,240,240,240,240,240,240,240,240,240,240,240,
  240,240,240,240,240,240,240,240,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,241,242,242,242,242,242,242,242,242,242,242,242,242,242,242,242,242,242,242,242,242,242,242,242,
  242,242,242,242,242,242,242,242,242,242,242,242,242,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,243,244,244,244,244,244,244,244,244,244,244,244,
  244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,244,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,
  245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,245,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,246,
  246,246,246,246,246,246,246,246,246,246,246,246,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,247,
  247,247,247,247,247,247,247,247,247,247,247,247,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,
  248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,248,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,
  249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,249,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,
  250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,250,
  250,250,250,250,250,250,250,250,250,250,250,250,250,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,
  251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,251,
  251,251,251,251,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,
  252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,
  252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,252,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,
  253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,
  253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,
  253,253,253,253,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,
  254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,
  254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,
  254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255
};

//~reciprocal look up table
// for (int i = 0; i < 128; i++) v = UINT16_MAX / (i + 1) - 1;
const int16_t rcp[128] = {
  -1, 32767, 21844, 16383, 13106, 10922, 9361, 8191, 7281, 6553, 5957, 5460,
  5040, 4680, 4368, 4095, 3854, 3640, 3448, 3276, 3120, 2978, 2848, 2730,
  2620, 2520, 2426, 2340, 2259, 2184, 2113, 2047, 1985, 1927, 1871, 1819,
  1770, 1724, 1679, 1637, 1597, 1559, 1523, 1488, 1455, 1424, 1393, 1364,
  1336, 1310, 1284, 1259, 1236, 1213, 1191, 1169, 1149, 1129, 1110, 1091,
  1073, 1056, 1039, 1023, 1007, 992, 977, 963, 949, 935, 922, 909,
  897, 885, 873, 861, 850, 839, 829, 818, 808, 798, 789, 779,
  770, 761, 752, 744, 735, 727, 719, 711, 704, 696, 689, 682,
  675, 668, 661, 654, 648, 642, 635, 629, 623, 617, 611, 606,
  600, 595, 589, 584, 579, 574, 569, 564, 559, 554, 550, 545,
  541, 536, 532, 528, 523, 519, 515, 511
};

int8_t noteMap[_MAX_NOTES][128];

class Note {
 public:
  uint8_t channel;
  uint8_t note;
  int8_t state;
  // fixed point precalculated parameters
  uf8p24 phase;
  uf8p24 freqInc;
  uf8p24 envelopePhase;
  uf0p16 envelopePhaseIncA;
  uf0p16 envelopePhaseIncD;
  uf0p16 envelopePhaseIncR;
  uf0p8 envelopeAmpS;
  uf0p8 envelopeAmpR;
  uf0p8 envelopeAmp;
  uf0p8 envelopeAmpDeltaS;
  uf0p8 envelopeAmpDeltaA;
  uf0p8 velocity;
  uf0p8 envelopeAmpA;

  Note() {
    state = -1;
    envelopeAmp = 0;
    memset (&noteMap, -1, sizeof (noteMap));
  }

  // v:0..127
  void setEnvelope (uint8_t v) {
    velocity = (v << 1) + 1;     // scale velocity to [1 - 255]
    // precalculate values
    // each envelope phase counts for 65536 steps (16 bit). the actual time they take is resulting from the increment we count up to thes value
    // the upper 8 bits of the envelope phase indicate in which region of ADSR we are
    envelopePhase = 0;
    envelopePhaseIncA = (rcp[adsr[0]]);                                    // 1/A
    envelopePhaseIncD = (rcp[adsr[1]]);                                    // 1/D
    envelopePhaseIncR = (rcp[adsr[3]]);                                    // 1/R
    envelopeAmpS = mul_uf0p8_uf0p8_uf0p8(velocity, (adsr[2] << 1) + 1); // S (scaled to 1 - 255) * velocity
    envelopeAmpDeltaS = velocity - envelopeAmpS;       // delta between max amp
    envelopeAmpDeltaA = velocity - envelopeAmpA;       // delta from 0 or last amplitude if note is replayed
  }

  void replay (uint8_t v) {
    envelopeAmpA = envelopeAmp; // start from current amplitude to avoid clicks
    setEnvelope (v);     // write new parameters
  }

  void on (uint8_t ch, uint8_t n, uint8_t v) {
    // set note parameters
    channel = ch;
    note = n; 
    phase = 0;
    freqInc = freqIncTab[n];
    freqInc <<= 2;      // have to quadruple the increment because of the we went down to a qarter of the frequency
    envelopeAmpA = 0;   // no playing atm start attack from 0
    setEnvelope (v);
    state = 0;          // set statte to plaing
  }

  void off (void) {
    envelopeAmpR = envelopeAmp; // save release amplitude
    if(envelopeAmpR)
      envelopePhase = 0x3000000;   // goto release
    else
      envelopePhase = 0x4000000;   // no relase phase
  }

  int8_t mix (void) {
    switch(cast_uf8p24_uf8p0 (envelopePhase)) {  //upper 8bits of envelope pahse are taking to identify the phase
      case 0:
        envelopeAmp = envelopeAmpA + mul_uf0p8_uf0p8_uf0p8(cast_uf8p24_uf0p8(envelopePhase), envelopeAmpDeltaA);  //ramping up during Attack pahse to velocity amp
        envelopePhase += envelopePhaseIncA;  //adding the increment. the higher the inc the shorter the attack period
        break;
      case 1:
        envelopeAmp = velocity - mul_uf0p8_uf0p8_uf0p8(cast_uf8p24_uf0p8(envelopePhase), envelopeAmpDeltaS);  //ramping decay down to sustain level
        envelopePhase += envelopePhaseIncD;  //same procedure her as in attack
        break;
      case 2:
        if(!envelopeAmpS) { //if sustain level is zerowe are done here, go to end
          envelopePhase = 0x4000000;
          return 0;
        }
        envelopeAmp = envelopeAmpS;      //set the amplitude to calculated value othervise
        break;
      case 3:
        envelopeAmp = envelopeAmpR - mul_uf0p8_uf0p8_uf0p8(cast_uf8p24_uf0p8(envelopePhase), envelopeAmpR);  //ramp down the last aplitude we saved to zero during the relese pahse 
        envelopePhase += envelopePhaseIncR;  //same as above
        break;
      case 4:
        state = -1;              //done with the note, clear the states
        noteMap[channel][note] = -1;
        return 0;
    }
    f8p0 s = 0;
    s = wave[cast_uf8p24_uf8p0(phase)];  //read the waveform from the table taking the high 8 bits of the time phase
    s = mul_f8p0_uf0p8_f8p0 (s, envelopeAmp);  //mutliplying the wave by the amplutide calculated from the ADSR and velocity above
    phase += freqInc;  //increment the pahase of the wave freqency 
    return s;
  }
};

Note notes[_MAX_NOTES];

void noteOff (uint8_t ch, uint8_t n, uint8_t v)  {
  if (noteMap[ch][n] >= 0) notes[noteMap[ch][n]].off ();
}

void noteOn (uint8_t ch, uint8_t n, uint8_t v) {
  if (!v) noteOff (ch, n, 0);
  else {
    if (noteMap[ch][n] >= 0) {
      notes[noteMap[ch][n]].replay (v);
    } else {
      for (uint8_t i = 0; i < _MAX_NOTES; i++) {
        if(notes[i].state == -1) {
          notes[i].on (ch, n, v);
          noteMap[ch][n] = i;
          return;
        }
      }
    }
  }
}

// Set Envelope
// V[0]:Attack, V[1]:Decay, V[2]:Sustain, V[3]Release
// range:0..127
void setEnvelope (const int8_t v[4]) {
  for (int i = 0; i < 4; i++) adsr[i] = v[i];
}

// Set Wave form
// v:0..127
void setWaveForm (const int8_t v[8]) {
  for (uint16_t i = 0; i < 256; i++) {  //interpolate the new wave form
    uint8_t par = i >> 5;
    uint8_t phase = i & 31;
    int16_t swave = (v[par] << 1) + (((v[(par + 1) & 7] - v[par]) * phase) >> 4);  // linear interpoaltion using some fixed point between this control point and the next
    wave[i] = swave - 128;  // store a signed sample
  }
}

// DAC
//-----------------------------
void DAC_Init (void) {
  const TPin p = { _GPIO9, PIO_TYPE_FIXED, SWM_FIXED_DAC_0, PIO_MODE_DEFAULT }; // DAC0
  PIO_Configure (&p, 1);
  // DACはこちらの設定も必要
  LPC_IOCON->PIO0_17 |= IOCON_PIO_DACMODE_M;

  // DACのクロックを活性化
  Chip_DAC0_Init ();
  // 割り込み及びDMAを不使用
  Chip_DAC_ConfigDMAConverterControl (LPC_DAC0, 0);
  Chip_DAC_SetDMATimeOut (LPC_DAC0, 0);
}

// MRT
//-----------------------------
// 指定周期毎にDACの出力を更新
extern "C" void MRT_IRQHandler (void) {
  Chip_MRT_IntClear (LPC_MRT_CH0);

  static int16_t v = 511;

  // DAC出力更新
  Chip_DAC_UpdateValue (LPC_DAC0, v);

  // 各ノートの合成
  v = 0;
  for (int8_t i = 0; i < _MAX_NOTES; i++)
    if (notes[i].state >= 0) v += notes[i].mix ();

  // PCMの合成
  for (uint8_t i = 0; i < (sizeof(PCMInfo) / sizeof(PCMInfo[0])); i++) {
    if (PCMInfo[i].ind != -1) {
      PCMInfo[i].ind += PCMInfo[i].freq;
      if ((PCMInfo[i].ind >> 8) >= PCMInfo[i].size) PCMInfo[i].ind = -1;
      else v += ((PCMInfo[i].raw[PCMInfo[i].ind >> 8]) << 2) * PCMInfo[i].velo / 127;
    }
  }

  v = compressTab[v + 2048] << 2; // 8bit->10bitへ

  // LEDの輝度更新
  LED_SetBright (abs (v - 512) << 1);
}

void MRT_Init (void) {
  Chip_MRT_Init ();

  Chip_MRT_SetInterval (LPC_MRT_CH0, Chip_Clock_GetSystemClockRate () / 11025);
  Chip_MRT_SetEnabled (LPC_MRT_CH0);
  Chip_MRT_SetMode (LPC_MRT_CH0, MRT_MODE_REPEAT);

  NVIC_EnableIRQ (MRT_IRQn);
}

// キー入力処理
//-----------------------------
// エンベロープの更新と表示
void SetEnvelope_disp (int16_t *d) {
  us_puts ("\rEnvelope{");
  for (int i = 0; i < 4; i++) {
    if (d[i] > 127) d[i] = 0;
    if (d[i] < 0) d[i] = 127;
    us_printf (" %d%c",d[i],i!=3?',':' ');
  }
  int8_t _d[4];
  for (int i = 0; i < 4; i++) _d[i] = d[i];
  setEnvelope (_d);
  us_puts ("}                   \r");
}

// ウェーブテーブルの更新と表示
void SetWave_disp (int16_t *d) {
  int8_t _d[8];
  us_puts ("\rWave{");
  for (int i = 0; i < 8; i++) {
    if (d[i] > 127) d[i] = 0;
    if (d[i] < 0) d[i] = 127;
    _d[i] = d[i];
    us_printf (" %d%c",d[i],i!=7?',':' ');
  }
  setWaveForm (_d);
  us_puts ("}                   \r");
}

void key (void) {
  int8_t oct = 4, pcm_tone = 5;
  bool pcmm = false;
  int16_t ADSR[4] = { 1, 16, 0, 0 }, WAVE[8] = { 100, 100, 100, 100, 28, 28, 28, 28 };

  SetEnvelope_disp (ADSR);
  us_puts ("\n");
  SetWave_disp (WAVE);
  us_puts ("\n");

  for (;;) {
    while (us_rx_buff ()) {
      char ch = us_getc ();
      switch (ch) {
        // 終了
        case '!':
          NVIC_SystemReset ();
          break;

        // エンベロープ変更
        case 0x22:ADSR[0]++; SetEnvelope_disp (ADSR); break;
        case 'W':ADSR[0]--; SetEnvelope_disp (ADSR); break;
        case '#':ADSR[1]++; SetEnvelope_disp (ADSR); break;
        case 'E':ADSR[1]--; SetEnvelope_disp (ADSR); break;
        case '$':ADSR[2]++; SetEnvelope_disp (ADSR); break;
        case 'R':ADSR[2]--; SetEnvelope_disp (ADSR); break;
        case '%':ADSR[3]++; SetEnvelope_disp (ADSR); break;
        case 'T':ADSR[3]--; SetEnvelope_disp (ADSR); break;

        // 波形編集
        case '1':WAVE[0]++; SetWave_disp (WAVE); break;
        case 'q':WAVE[0]--; SetWave_disp (WAVE); break;
        case '2':WAVE[1]++; SetWave_disp (WAVE); break;
        case 'w':WAVE[1]--; SetWave_disp (WAVE); break;
        case '3':WAVE[2]++; SetWave_disp (WAVE); break;
        case 'e':WAVE[2]--; SetWave_disp (WAVE); break;
        case '4':WAVE[3]++; SetWave_disp (WAVE); break;
        case 'r':WAVE[3]--; SetWave_disp (WAVE); break;
        case '5':WAVE[4]++; SetWave_disp (WAVE); break;
        case 't':WAVE[4]--; SetWave_disp (WAVE); break;
        case '6':WAVE[5]++; SetWave_disp (WAVE); break;
        case 'y':WAVE[5]--; SetWave_disp (WAVE); break;
        case '7':WAVE[6]++; SetWave_disp (WAVE); break;
        case 'u':WAVE[6]--; SetWave_disp (WAVE); break;
        case '8':WAVE[7]++; SetWave_disp (WAVE); break;
        case 'i':WAVE[7]--; SetWave_disp (WAVE); break;

        // オクターブ変更
        case 'o':if (oct < 9) oct++; us_printf ("\noct=%d", oct); break;
        case 'p':if (oct > 0) oct--; us_printf ("\noct=%d", oct); break;

        // 鍵盤 (小文字がノートON、大文字がノートOFF)
        // c
        case 'z': if (pcmm) PCM_play (pcm_tone, oct * 12 + 0, 64); else noteOn (0, oct * 12 + 0, 127); break;
        case 'Z': noteOff (0, oct * 12 + 0, 127); break;
        // c+
        case 's': if (pcmm) PCM_play (pcm_tone, oct * 12 + 1, 64); else noteOn (0, oct * 12 + 1, 127); break;
        case 'S': noteOff (0, oct * 12 + 1, 127); break;
        // d
        case 'x': if (pcmm) PCM_play (pcm_tone, oct * 12 + 2, 64); else noteOn (0, oct * 12 + 2, 127); break;
        case 'X': noteOff (0, oct * 12 + 2, 127); break;
        // d+
        case 'd': if (pcmm) PCM_play (pcm_tone, oct * 12 + 3, 64); else noteOn (0, oct * 12 + 3, 127); break;
        case 'D': noteOff (0, oct * 12 + 3, 127); break;
        // e
        case 'c': if (pcmm) PCM_play (pcm_tone, oct * 12 + 4, 64); else noteOn (0, oct * 12 + 4, 127); break;
        case 'C': noteOff (0, oct * 12 + 4, 127); break;
        // f
        case 'v': if (pcmm) PCM_play (pcm_tone, oct * 12 + 5, 64); else noteOn (0, oct * 12 + 5, 127); break;
        case 'V': noteOff (0, oct * 12 + 5, 127); break;
        // f+
        case 'g': if (pcmm) PCM_play (pcm_tone, oct * 12 + 6, 64); else noteOn (0, oct * 12 + 6, 127); break;
        case 'G': noteOff (0, oct * 12 + 6, 127); break;
        // g
        case 'b': if (pcmm) PCM_play (pcm_tone, oct * 12 + 7, 64); else noteOn (0, oct * 12 + 7, 127); break;
        case 'B': noteOff (0, oct * 12 + 7, 127); break;
        // g+
        case 'h': if (pcmm) PCM_play (pcm_tone, oct * 12 + 8, 64); else noteOn (0, oct * 12 + 8, 127); break;
        case 'H': noteOff (0, oct * 12 + 8, 127); break;
        // a
        case 'n': if (pcmm) PCM_play (pcm_tone, oct * 12 + 9, 64); else noteOn (0, oct * 12 + 9, 127); break;
        case 'N': noteOff (0, oct * 12 + 9, 127); break;
        // a+
        case 'j': if (pcmm) PCM_play (pcm_tone, oct * 12 + 10, 64); else noteOn (0, oct * 12 + 10, 127); break;
        case 'J': noteOff (0, oct * 12 + 10, 127); break;
        // b
        case 'm': if (pcmm) PCM_play (pcm_tone, oct * 12 + 11, 64); else noteOn (0, oct * 12 + 11, 127); break;
        case 'M': noteOff (0, oct * 12 + 11, 127); break;
        // <c
        case ',': if (pcmm) PCM_play (pcm_tone, oct * 12 + 12, 64); else noteOn (0, oct * 12 + 12, 127); break;
        case '<': noteOff (0, oct * 12 + 12, 127); break;
        // <c+
        case 'l': if (pcmm) PCM_play (pcm_tone, oct * 12 + 13, 64); else noteOn (0, oct * 12 + 13, 127); break;
        case 'L': noteOff (0, oct * 12 + 13, 127); break;
        // <d
        case '.': if (pcmm) PCM_play (pcm_tone, oct * 12 + 14, 64); else noteOn (0, oct * 12 + 14, 127); break;
        case '>': noteOff (0, oct * 12 + 14, 127); break;

        // 鍵盤で再生するソースをウェーブテーブルとPCMから選択
        case ';':
          pcmm = !pcmm;
          us_printf ("\nPCM mode %s", pcmm?"on":"off");
          for (int i = 0; i < _MAX_NOTES; i++) {
            notes[i].state = -1;
            notes[i].envelopeAmp = 0;
          }
          memset (&noteMap, -1, sizeof (noteMap));
          break;

        // PCM単体で再生
        case '^':
          PCM_play (0, 48, 64);
          break;
        case '\\':
          PCM_play (1, 48, 64);
          break;
        case '[':
          PCM_play (2, 48, 64);
          break;
        case ']':
          PCM_play (3, 48, 64);
          break;
        case '/':
          PCM_play (4, 48, 64);
          break;
        case ':':
          PCM_play (5, 48, 64);
          break;

        // ウェーブテーブル場合は最大分解能である256ステップで4タイプから選択
        // PCMの場合は6つの音色から選択
        case '@':
          char c;
          us_printf (pcmm?"\ninput 0..5:":"\ninput 0..3:");
          us_printf ("%c\n",c = us_getc ());

          if (pcmm) {
            switch (c) {
              case '0' ... '5':
                pcm_tone = c - '0';
                break;
            }
          } else {
            switch (c) {
              case '0':
                // ノコギリ
                for (int t = 0; t < 256; t++) wave[t] = ((t % 256) - 128) * 0.5;
                break;
              case '1':
                // 正弦波
                for (int t = 0; t < 256; t++) wave[t] = ((__builtin_sin ((t % 256) * M_PI * 2.0 / 255.0) * 255.0) / 2) * 0.5;
                break;
              case '2':
                // 矩形波
                for (int t = 0; t < 256; t++) wave[t] = ((((t % 256) < 128) * 255) - 128) * 0.5;
                break;
              case '3':
                // 三角波
                for (int t = 0; t < 256; t++) wave[t] = ((abs (t % 256 - 127) * 255) / 128 - 128) * 0.5;
                break;

            }
            for (int k = 0; k < 256; k++) us_printf("%s%d%s", (k==0)?"{":"",wave[k],(k!=255)?",":"}\n");
          }
          break;
      }
    }
    __WFI ();
  }
}

// main
//-----------------------------
int main (void) {
  // USART初期化
  static uint8_t txb[100], rxb[10];
  us_init (Chip_Clock_GetFRGClockRate (0), 115200, USIP_8N1, txb, sizeof (txb), rxb, sizeof (rxb));

  LED_Init ();
  DAC_Init ();
  MRT_Init ();

  us_puts ("\x1b[2JSTART\n");

  key ();
}
