﻿/*
  DYNAMIXELクライアント

  DYNAMIXELシリーズと同じネットワークに配置してクライアントとして機能
  させる。

  ・GPIOを個別に任意のモードに設定可能
  ・IMUを常時モニタしコントロールテーブルへ逐次反映
  ・キャプチャ及びカウンタ機能はペンディング
  ・旧DXMIOよりもメモリが増えた恩恵で、以前よりも多少凝ったユーザー任
    意タスクを追加可能
  ・armコアが2個載っている事になるので、フルパワーで動かすと少々発熱す
    る
 */

// DXプロトコルバージョン選択 (1 or 2)
#define DX_PROTOCOL_VER 2

#include  <FreeRTOS.h>
#include  <task.h>

#include  <chip.h>
// ボードの端子割り当てと最小限の初期化
#include  "e097b_def.h"
// IMU用サブルーチン
#include  "e097b_imu.h"
#include  <us.h>
// DXクライアントライブラリ
#if DX_PROTOCOL_VER == 1
#include  <dxclient.h>
#elif DX_PROTOCOL_VER == 2
#include  <dx2client.h>
#endif
#include  <wl_delay.h>
#include  <wl_adc.h>
#include  <wl_i2c.h>

// マクロと変数
//-----------------------------
//#define _USE_MYTASK

// DXMIO用Pin Mode
typedef enum { tPinDIN = 0, tPinDIN_PU, tPinDIN_PD, tPinDOUT, tPinDOUT_OC, tPinAIN, tPinPWM0, tPinPWM1, tPinPWM2, tPinPWM3, tPinPWM4, tPinPWM5, tPinMPW0, tPinMPW1, tPinMPW2, tPinMPW3, tPinMPC0, tPinMPC1, tPinMPC2, tPinMPC3, tPinAOUT, tPinSPFUNC } TMIOPinMode;

// Pin information
typedef struct {
  uint8_t               No;
  CHIP_PINx_T           CPin;
  CHIP_SWM_PIN_FIXED_T  AD;
} TPinInfo;

// 外部端子
const TPinInfo PinInfo[12] = {
  { _GPIO0,  IOCON_PIO0_7,  SWM_FIXED_ADC_0  },
  { _GPIO1,  IOCON_PIO0_6,  SWM_FIXED_ADC_1  },
  { _GPIO2,  IOCON_PIO0_14, SWM_FIXED_ADC_2  },
  { _GPIO3,  IOCON_PIO0_23, SWM_FIXED_ADC_3  },
  { _GPIO4,  IOCON_PIO0_22, SWM_FIXED_ADC_4  },
  { _GPIO5,  IOCON_PIO0_21, SWM_FIXED_ADC_5  },
  { _GPIO6,  IOCON_PIO0_20, SWM_FIXED_ADC_6  },
  { _GPIO7,  IOCON_PIO0_19, SWM_FIXED_ADC_7  },
  { _GPIO8,  IOCON_PIO0_18, SWM_FIXED_ADC_8  },
  { _GPIO9,  IOCON_PIO0_17, SWM_FIXED_ADC_9  },
  { _GPIO10, IOCON_PIO0_13, SWM_FIXED_ADC_10 },
  { _GPIO11, IOCON_PIO0_4,  SWM_FIXED_ADC_11 }
};

TMIOPinMode previous_setting[12] = { tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN };

TOperationMode imu_opmode = opmode_NDOF;
TAxisPlace imu_axisplace = axisplace_P0;

// シリアル通信バッファ
static uint8_t txd[200], rxd[250];

// コントロールテーブル
typedef struct {
  // V1:0x00/V2:0x0000
  uint16_t  ModelNumber;      // モデルナンバー (R)
#if DX_PROTOCOL_VER == 2
  uint32_t  ModelInfomation;  // モデル情報 (R)
#endif
  uint8_t   FirmwareVersion;  // ファームウェアバージョン (R)
  uint8_t   ID;               // ID (R/W & NVM)
  uint8_t   Baudrate;         // ボーレート (R/W & NVM)
  uint8_t   WriteNVM;         // 不揮発データ書き込みフラグ (R/W)

  uint8_t   LED;              // LED2  (R/W)
  uint8_t   Terminator;       // 終端抵抗 (R/W & NVM)
  uint8_t   PinCfg[12];       // GPIO端子機能設定 (R/W & NVM)
  uint16_t  PWMFrequency;     // PWM周波数設定 (R/W & NVM)
  uint16_t  PWMDuty[6];       // PWMデューティー設定 (R/W & NVM)
  uint8_t   reserve2[10];
  uint16_t  Capture[4];       // パルス設定・計測値 (R/W)
  uint16_t  OUT;              // GPIO出力設定値 (R/W)
  uint16_t  IN;               // GPIO入力値 (R)
  uint16_t  ADV[12];          // アナログ電圧測定値 (R)
  uint16_t  DAV[12];          // アナログ電圧設定値 (R/W)
  uint8_t   IMU_OprMode;      // IMU動作モード (R/W & NVM)
  uint8_t   IMU_AxisPlace;    // IMUセンサ配置 (R/W & NVM)
  Tbno055   IMU;              // IMU測定値 (R)
  uint8_t   IMU_stat;         // IMUステータス値 (R)
  uint8_t   IMU_err;          // IMUエラー値 (R)
  uint8_t   userarea[20];     //
} __attribute__ ((packed)) TE097BCtrlTable;

// コントロールテーブルのポインタ
TE097BCtrlTable  *pmemmap;

// コントロールテーブルの各アイテムのオフセットアドレス
#define _ADDR_PinCfg        ((int)&((TE097BCtrlTable*)0)->PinCfg)
#define _ADDR_PWMFrequency  ((int)&((TE097BCtrlTable*)0)->PWMFrequency)
#define _ADDR_PWMDuty       ((int)&((TE097BCtrlTable*)0)->PWMDuty)
#define _ADDR_Capture       ((int)&((TE097BCtrlTable*)0)->Capture)
#define _ADDR_OUT           ((int)&((TE097BCtrlTable*)0)->OUT)
#define _ADDR_IN            ((int)&((TE097BCtrlTable*)0)->IN)
#define _ADDR_ADV           ((int)&((TE097BCtrlTable*)0)->ADV)
#define _ADDR_IMU_OprMode   ((int)&((TE097BCtrlTable*)0)->IMU_OprMode)
#define _ADDR_IMU           ((int)&((TE097BCtrlTable*)0)->IMU)
#define _ADDR_IMU_stat      ((int)&((TE097BCtrlTable*)0)->IMU_stat)
#define _ADDR_IMU_err       ((int)&((TE097BCtrlTable*)0)->IMU_err)
#define _ADDR_userarea      ((int)&((TE097BCtrlTable*)0)->userarea)

// コントロールテーブルデフォルト値
static const TE097BCtrlTable DEFAULT_DATA = {
  ModelNumber     : 0x4023,
#if DX_PROTOCOL_VER == 2
  ModelInfomation : 0,
#endif
  FirmwareVersion : 'A',
  ID              : 200,
#if DX_PROTOCOL_VER == 1
  Baudrate        : 1,
#elif DX_PROTOCOL_VER == 2
  Baudrate        : 3,
#endif
  WriteNVM        : 0,
  LED             : 0,
  Terminator      : 1,

  IMU_OprMode     : opmode_NDOF,
  IMU_AxisPlace   : axisplace_P0,
};

// PWM
//-----------------------------
static void init_SCT (void) {
  static bool inited = false;

  if (!inited) {
    Chip_SCT_Init (LPC_SCT);
    inited = true;
  }
}

static void update_PWMFreqency (uint16_t freq) {
  static uint16_t oldfreq = 0;

  if (freq == 0) {
    Chip_SCTPWM_Stop (LPC_SCT);
    Chip_SCT_DeInit (LPC_SCT);
  } else {
    init_SCT();
    if (oldfreq != freq) Chip_SCTPWM_SetRate (LPC_SCT, freq);
  }
  oldfreq = freq;
}

static void update_PWMDuty (uint8_t ch, uint16_t duty) {
  uint32_t d;

  Chip_SCTPWM_SetOutPin (LPC_SCT, ch + 1, ch);
  if (duty == 0)          d = 0;
  else if (duty == 65535) d = Chip_SCTPWM_GetTicksPerCycle (LPC_SCT) + 1;
  else                    d = ((uint64_t)duty * (uint64_t)Chip_SCTPWM_GetTicksPerCycle (LPC_SCT)) / (uint64_t) (65535UL);
  Chip_SCTPWM_SetDutyCycle (LPC_SCT, ch + 1, d);
  Chip_SCTPWM_Start (LPC_SCT);
}

// Capture
//-----------------------------
#if 0
uint16_t pwd[4];

/**
 * @brief SCT Input Mux Index; See Chip_INMUX_SetSCTInMux()
 */
typedef enum {
  SCT_INMUX_0,   /*!< Input mux for SCT0; INPUT 0 */
  SCT_INMUX_1,   /*!< Input mux for SCT0; INPUT 1 */
  SCT_INMUX_2,   /*!< Input mux for SCT0; INPUT 2 */
  SCT_INMUX_3,   /*!< Input mux for SCT0; INPUT 3 */
} SCT_INMUX_T;

/**
 * @brief SCT INPUT triggers
 */
typedef enum {
  SCT_INP_IN0,                /*!< SCT0_IN0 selected by Pin Matrix */ /* FIXME: UM hints about changes */
  SCT_INP_IN1,                /*!< SCT0_IN1 selected by Pin Matrix */
  SCT_INP_IN2,                /*!< SCT0_IN2 selected by Pin Matrix */
  SCT_INP_IN3,                /*!< SCT0_IN3 selected by Pin Matrix */
  SCT_INP_ADC_THCMP_IRQ,      /*!< ADC Threshold compare IRQ */
  SCT_INP_ACMP_O,             /*!< Analog comparator output */
  SCT_INP_ARM_TXEV,           /*!< ARM TX Event */
  SCT_INP_DEBUG_HALTED,       /*!< Debug halted event */
} SCT_INP_T;

/**
 * @brief Select a trigger source SCT module
 * @param pINMUX  : The base of INPUT MUX register block
 * @param isct  : Index of SCT input mux
 * @param trig  : SCT Input function that will cause the trigger
 * @return  Nothing
 */
STATIC INLINE void Chip_INMUX_SetSCTInMux (LPC_INMUX_INMUX_T *pINMUX, SCT_INMUX_T isct, SCT_INP_T trig) {
  pINMUX->SCT_INMUX[isct] = trig;
}

static void init_SCTCapture (void) {
  init_SCT();

#if 0
  /* Initialize the SCT clock and reset the SCT */
  Chip_Clock_EnablePeriphClock (SYSCTL_CLOCK_SCT);
  Chip_SYSCTL_PeriphReset (RESET_SCT);
  Chip_INMUX_SetSCTInMux (LPC_INMUX, SCT_INMUX_0, SCT_INP_IN0);

  /* Enable flag to request an interrupt for Event 0 */
  Chip_SCT_EnableEventInt (LPC_SCT, SCT_EVT_0);

  /* Enable the interrupt for the SCT */
  NVIC_EnableIRQ (SCT_IRQn);

  /* Start the SCT counter by clearing Halt_L in the SCT control register */
  Chip_SCT_ClearControl (LPC_SCT, SCT_CTRL_HALT_L);


  LPC_SCT->CONFIG |= 0x1;                            // unified timer

  LPC_SCT->REGMODE_U = 0x1;                           //match0 operates as capture register
  LPC_SCT->CAPCTRL[0].U = 0x1;                          //event 0 causes capture 0

  LPC_SCT->EV[0].STATE = 0x00000001;                  // ev 0 happens in state 0
  LPC_SCT->EV[0].CTRL  = (1 << 0)  |                  // related to match 0
                         (0 << 5) |             //select input
                         (0 << 6) |             //select sct_input0
                         (2 << 10) |              //falling edge
                         (2 << 12) |              // input condition only
                         (1 << 14) |              // STATEV is new state
                         (0 << 15);               // STATEV[15] = 0
#else
//#define PWM_FREQUENCY 10000 // PWM frequency in Hz
//#define PWM_RESOLUTION_NS 1000 // Timer resolution in ns
#define PWM_FREQUENCY 50        // PWM frequency in Hz
#define PWM_RESOLUTION_NS 100  // Timer resolution in ns
#define PWM_MIN_DUTY_PERCENT 5  // Minimum allowed duty cycle in %
#define PWM_MAX_DUTY_PERCENT 80 // Maximum allowed duty cycle in %

  /* Derived constants */
#define SCT_PRESCALER           (((SystemCoreClock / 1000u) * PWM_RESOLUTION_NS) / 1000000u - 1u)

#define match_min_width         ((10000000u * PWM_MIN_DUTY_PERCENT) / (PWM_FREQUENCY * PWM_RESOLUTION_NS))
#define match_max_width         ((10000000u * PWM_MAX_DUTY_PERCENT) / (PWM_FREQUENCY * PWM_RESOLUTION_NS))
#define match_no_input          ((10000000u * 300                 ) / (PWM_FREQUENCY * PWM_RESOLUTION_NS))

  Chip_INMUX_SetSCTInMux (LPC_INMUX_INMUX, SCT_INMUX_0, SCT_INP_IN0);

  LPC_SCT->CONFIG |= (1 << 0) | (1 << 17); // unified, auto limit
  LPC_SCT->CTRL_U |= (SCT_PRESCALER << 5); // set prescaler
  LPC_SCT->REGMODE_L = 0x00000018; // 3x MATCH, 2x CAPTURE used
  LPC_SCT->MATCH[0].U = match_max_width; // match_max_width
  LPC_SCT->MATCHREL[0].U = match_max_width;
  LPC_SCT->MATCH[1].U = match_min_width; // match_min_width
  LPC_SCT->MATCHREL[1].U = match_min_width;
  LPC_SCT->MATCH[2].U = match_no_input; // match_no_input
  LPC_SCT->MATCHREL[2].U = match_no_input;
  LPC_SCT->EVENT[0].STATE = 0xFFFFFFFF; // event 0 happens in all states
  LPC_SCT->EVENT[0].CTRL = (2 << 0) | (1 << 12); // related to match_no_input only
  LPC_SCT->EVENT[1].STATE = 0xFFFFFFFF; // event 1 happens in all states
  LPC_SCT->EVENT[1].CTRL = (1 << 10) | (2 << 12); // IN_0 rising edge condition only
  LPC_SCT->EVENT[2].STATE = (1 << 0); // event 2 happens in state 0
  LPC_SCT->EVENT[2].CTRL = (1 << 0) | // related to match_min_width
                           (1 << 12) | // match condition only
                           (1 << 14) | // STATEV is loaded into state
                           (1 << 15); // new state is 1
  LPC_SCT->EVENT[3].STATE = (1 << 1); // event 3 happens in state 1
  LPC_SCT->EVENT[3].CTRL = (2 << 10) | // IN_0 falling edge
                           (2 << 12) | // IO condition only
                           (1 << 14) | // STATEV is loaded into state
                           (0 << 15); // new state is 0
  LPC_SCT->EVENT[4].STATE = (1 << 1); // event 4 happens in state 1
  LPC_SCT->EVENT[4].CTRL = (0 << 0) | // related to match_max_width
                           (1 << 12) | // match condition only
                           (1 << 14) | // STATEV is loaded into state
                           (0 << 15); // new state is 0
  LPC_SCT->EVENT[5].STATE = (1 << 0); // event 5 happens in state 0
  LPC_SCT->EVENT[5].CTRL = (2 << 10) | (2 << 12); // IN_0 falling edge condition only
  LPC_SCT->CAPCTRL[3].U = (1 << 1); // event 1 is causing capture 3
  LPC_SCT->CAPCTRL[4].U = (1 << 3) | (1 << 5); // event 3 and 5 cause capture 4
  LPC_SCT->OUT[0].SET = (1 << 1); // event 1 set OUT0 (no timeout)
  LPC_SCT->OUT[0].CLR = (1 << 0); // event 0 clear OUT0 (timeout)
  LPC_SCT->OUT[1].SET = (1 << 3); // event 3 set OUT1 (no width error)
  LPC_SCT->OUT[1].CLR = (1 << 0) | (1 << 5); // event 0 and 5 clear OUT1 (width error)
  LPC_SCT->OUTPUT |= 3; // default set OUT0 and OUT1
  LPC_SCT->LIMIT_L = (1 << 0) | (1 << 1); // event 0 and 1 limit the timer
  LPC_SCT->EVEN = (1 << 0) | (1 << 5); // event 0 and 5 generate an irq
  NVIC_EnableIRQ (SCT_IRQn); // enable SCTimer/PWM interrupt
  LPC_SCT->CTRL_U &= ~ (1 << 2); // start timer
#endif
}

void SCT_IRQHandler (void) {
#if 0
//  cap_flag = 1;
  volatile static uint32_t oldcap = 0;
  uint32_t cap = LPC_SCT->CAP[0].U, n;

//  if (cap >= oldcap) {
  n = cap - oldcap;
  pwd[0] = n;
  pwd[1] = n >> 16;
//  } else {
//    n = (UINT32_MAX - oldcap) + cap;
//    pwd[0] = n;
//    pwd[1] = n >> 16;
//  }
  oldcap = cap;
  /* Clear the SCT Event 0 Interrupt */
  Chip_SCT_ClearEventFlag (LPC_SCT, SCT_EVT_0);
#else
#define SCT_EVT_5                 (1 << 5)

  uint32_t status = LPC_SCT->EVEN;

  if (status & SCT_EVT0_BIT) {              // EVENT 0: timeout
    /* New measurement result */
    pwd[2] = 0x0400;
    ;
  }
  if (status & SCT_EVT5_BIT) {              // EVENT 5: width error
    /* Time-out (no signal) */
    ;
    pwd[2] = 0x8000;
  }

  LPC_SCT->EVFLAG = status;             // Acknowledge interrupts
#endif
}
#endif

// pin configuration
//-----------------------------
static void pin_config (uint8_t adch, TMIOPinMode pm, uint16_t freq, uint16_t duty) {
  static const uint8_t SCT_OUT_pin_movable[6] = { SWM_SCTOUT_0_O, SWM_SCTOUT_1_O, SWM_SCTOUT_2_O, SWM_SCTOUT_3_O, SWM_SCTOUT_4_O, SWM_SCTOUT_5_O };
  static const uint8_t SCT_IN_pin_movable[4] = { SWM_SCTIN_0_I, SWM_SCTIN_1_I, SWM_SCTIN_2_I, SWM_SCTIN_3_I };
  TPin pin = { PinInfo[adch].No, PIO_TYPE_FIXED, 0, PIO_MODE_DEFAULT };
  uint8_t ch;

  // 割り当て済みSCTの切断
  if ((previous_setting[adch] >= tPinPWM0) && (previous_setting[adch] <= tPinPWM5)) {
    ch = previous_setting[adch] - tPinPWM0;
    Chip_Clock_EnablePeriphClock (SYSCON_CLOCK_SWM);
    Chip_SWM_MovablePinAssign ((CHIP_SWM_PIN_MOVABLE_T)SCT_OUT_pin_movable[ch], 0xff);
    Chip_Clock_DisablePeriphClock (SYSCON_CLOCK_SWM);
  }
  if ((previous_setting[adch] >= tPinMPW0) && (previous_setting[adch] <= tPinMPW3)) {
    ch = previous_setting[adch] - tPinMPW0;
    Chip_Clock_EnablePeriphClock (SYSCON_CLOCK_SWM);
    Chip_SWM_MovablePinAssign ((CHIP_SWM_PIN_MOVABLE_T)SCT_IN_pin_movable[ch], 0xff);
    Chip_Clock_DisablePeriphClock (SYSCON_CLOCK_SWM);
  }
  if ((previous_setting[adch] >= tPinMPC0) && (previous_setting[adch] <= tPinMPC3)) {
    ch = previous_setting[adch] - tPinMPC0;
    Chip_Clock_EnablePeriphClock (SYSCON_CLOCK_SWM);
    Chip_SWM_MovablePinAssign ((CHIP_SWM_PIN_MOVABLE_T)SCT_IN_pin_movable[ch], 0xff);
    Chip_Clock_DisablePeriphClock (SYSCON_CLOCK_SWM);
  }

  pin.gpio_no = PinInfo[adch].No;

  if (pin.gpio_no == _GPIO9) LPC_IOCON->PIO0_17 &= ~ (1 << IOCON_PIO_DACMODE_P);

  switch (pm) {
    case tPinDIN:     // 0
      PIO_Configure (&pin, 1);
      pin.pin_type = PIO_TYPE_INPUT;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinDIN_PU:  // 1
      pin.pin_type = PIO_TYPE_INPUT;
      pin.pin_mode = PIO_MODE_PULLUP;
      break;
    case tPinDIN_PD:  // 2
      pin.pin_type = PIO_TYPE_INPUT;
      pin.pin_mode = PIO_MODE_PULLDOWN;
      break;
    case tPinDOUT:    // 3
      pin.pin_type = PIO_TYPE_OUTPUT;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinDOUT_OC: // 4
      pin.pin_type = PIO_TYPE_OUTPUT;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinAIN:     // 5
      pin.pin_type = PIO_TYPE_FIXED;
      pin.pin_movable_fixed = PinInfo[adch].AD;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinPWM0:    // 6
    case tPinPWM1:    // 7
    case tPinPWM2:    // 8
    case tPinPWM3:    // 9
    case tPinPWM4:    // 10
    case tPinPWM5:    // 11
      ch = pm - tPinPWM0;
      // disconnect
      Chip_Clock_EnablePeriphClock (SYSCON_CLOCK_SWM);
      Chip_SWM_MovablePinAssign ((CHIP_SWM_PIN_MOVABLE_T)SCT_OUT_pin_movable[ch], 0xff);
      Chip_Clock_DisablePeriphClock (SYSCON_CLOCK_SWM);

      pin.pin_type = PIO_TYPE_MOVABLE;
      pin.pin_movable_fixed = SCT_OUT_pin_movable[ch];
      pin.pin_mode = PIO_MODE_DEFAULT;
      update_PWMFreqency (freq);
      update_PWMDuty (ch, duty);
      break;
    case tPinMPW0:    // 12
    case tPinMPW1:    // 13
    case tPinMPW2:    // 14
    case tPinMPW3:    // 15
      ch = pm - tPinMPW0;
      // disconnect
      Chip_Clock_EnablePeriphClock (SYSCON_CLOCK_SWM);
      Chip_SWM_MovablePinAssign ((CHIP_SWM_PIN_MOVABLE_T)SCT_IN_pin_movable[ch], 0xff);
      Chip_Clock_DisablePeriphClock (SYSCON_CLOCK_SWM);

      pin.pin_type = PIO_TYPE_MOVABLE;
      pin.pin_movable_fixed = SCT_IN_pin_movable[ch];
      pin.pin_mode = PIO_MODE_DEFAULT;
//      init_SCTCapture ();
      break;
    case tPinMPC0:    // 16
    case tPinMPC1:    // 17
    case tPinMPC2:    // 18
    case tPinMPC3:    // 19
      ch = pm - tPinMPC0;
      // disconnect
      Chip_Clock_EnablePeriphClock (SYSCON_CLOCK_SWM);
      Chip_SWM_MovablePinAssign ((CHIP_SWM_PIN_MOVABLE_T)SCT_IN_pin_movable[ch], 0xff);
      Chip_Clock_DisablePeriphClock (SYSCON_CLOCK_SWM);

      pin.pin_type = PIO_TYPE_MOVABLE;
      pin.pin_movable_fixed = SCT_IN_pin_movable[ch];
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinAOUT:    // 20
      // only pio0_17
      pin.pin_type = PIO_TYPE_FIXED;
      pin.pin_movable_fixed = SWM_FIXED_DAC_0;
      pin.pin_mode = PIO_MODE_DEFAULT;
      LPC_IOCON->PIO0_17 |= (1 << IOCON_PIO_DACMODE_P);
    case tPinSPFUNC:  // 21
      break;
  }
  previous_setting[adch] = pm;
  PIO_Configure (&pin, 1);
}

// WDT
//-----------------------------
static void init_wdt (void) {
  uint32_t wdtFreq;

  Chip_Clock_SetWDTOSC (WDTLFO_OSC_0_60, 64);     // Freq = 0.6Mhz, divided by 64. WDT_OSC should be 9.375khz
  Chip_SYSCON_PowerUp (SYSCON_PDRUNCFG_WDTOSC);   // Enable the power to the WDT
  wdtFreq = Chip_Clock_GetWDTOSCRate() / 4;       // The WDT divides the input frequency into it by 4
  Chip_WWDT_Init (LPC_WWDT);                      // Initialize WWDT (also enables WWDT clock)
  /* Set watchdog feed time constant to approximately 2s
     Set watchdog warning time to 512 ticks after feed time constant
     Set watchdog window time to 3s */
  Chip_WWDT_SetTimeOut (LPC_WWDT, wdtFreq * 2);
  Chip_WWDT_SetWarning (LPC_WWDT, 512);
  Chip_WWDT_SetWindow (LPC_WWDT, wdtFreq * 3);
  Chip_WWDT_SetOption (LPC_WWDT, WWDT_WDMOD_WDRESET);  // Configure WWDT to reset on timeout
  Chip_WWDT_ClearStatusFlag (LPC_WWDT, WWDT_WDMOD_WDTOF | WWDT_WDMOD_WDINT); // Clear watchdog warning and timeout interrupts
  Chip_WWDT_Start (LPC_WWDT);  // Start watchdog
}

// USART
//-----------------------------
#if DX_PROTOCOL_VER == 2
static const uint32_t baud[9] = { 9600, 57600, 115200, 1000000, 2000000, 3000000, 4000000, 4500000, 10500000 };
#endif

// 初期化
#if DX_PROTOCOL_VER == 1
static uint32_t us_init (uint32_t baudrate) {
  return usart0_dma_init (Chip_Clock_GetFRGClockRate (0), baudrate, USIP_8N1, txd, sizeof (txd), rxd, sizeof (rxd));
}
#elif DX_PROTOCOL_VER == 2
static uint32_t us_init (uint8_t ind) {
  if (ind > 6) ind = 6;
  else if (ind == 5) ind = 4;
  return usart0_dma_init (Chip_Clock_GetFRGClockRate (0), baud[ind], USIP_8N1, txd, sizeof (txd), rxd, sizeof (rxd));
}
#endif

// ボーレート更新
#if DX_PROTOCOL_VER == 1
static uint32_t us_setbaudrate (uint32_t baudrate) {
  return usart0_dma_setbaudrate (Chip_Clock_GetFRGClockRate (0), baudrate);
}
#elif DX_PROTOCOL_VER == 2
static uint32_t us_setbaudrate (uint8_t ind) {
  if (ind > 6) ind = 6;
  else if (ind == 5) ind = 4;
  return usart0_dma_setbaudrate (Chip_Clock_GetFRGClockRate (0), baud[ind]);
}
#endif

// 終了
static void us_deinit (void) { }

// DXクライアントコールバック
//-----------------------------
// アイテムのレンジチェック(個別)
#if DX_PROTOCOL_VER == 1
static bool DX_CallbackValidPrevMapRange (const void *pmem, uint8_t u_st, uint8_t u_len) {
#elif DX_PROTOCOL_VER == 2
static bool DX_CallbackValidPrevMapRange (const void *pmem, uint16_t u_st, uint16_t u_len) {
#endif
  bool result = true;
  const TE097BCtrlTable *pNewControlTable = (TE097BCtrlTable *)pmem;

  for (int i = 0; i < 12; i++) {
    if (IS_SAFE_VAR (pNewControlTable, pNewControlTable->PinCfg[i], u_st, u_len)) {
      if ((pNewControlTable->PinCfg[i] > tPinSPFUNC) || ((i != 9) && (pNewControlTable->PinCfg[i] == tPinAOUT))) {
        result = false;
        break;
      } else {
        for (int j = 0; j < 12; j++) {
          if ((i != j) && (pNewControlTable->PinCfg[i] >= tPinPWM0) && (pNewControlTable->PinCfg[i] <= tPinMPC3) && (pNewControlTable->PinCfg[i] == pmemmap->PinCfg[j])) {
            result = false;
            break;
          }
        }
      }
    }
  }
  if (IS_SAFE_VAR (pNewControlTable, pNewControlTable->PWMFrequency, u_st, u_len)) {
    if (pNewControlTable->PWMFrequency > 50000) result = false;
  }
  if (IS_SAFE_VAR (pNewControlTable, pNewControlTable->OUT, u_st, u_len)) {
    if (pNewControlTable->OUT > 0xfff) result = false;
  }
  if (IS_SAFE_VAR (pNewControlTable, pNewControlTable->OUT, u_st, u_len)) {
    if (pNewControlTable->OUT > 0xfff) result = false;
  }
  if (IS_SAFE_VAR (pNewControlTable, pNewControlTable->IMU_OprMode, u_st, u_len)) {
    if (pNewControlTable->IMU_OprMode > 12) result = false;
  }

#if DX_PROTOCOL_VER == 2
  if (IS_SAFE_VAR (pNewControlTable, pNewControlTable->Baudrate, u_st, u_len)) {
    if (! ((pNewControlTable->Baudrate <= 4) || (pNewControlTable->Baudrate == 6))) result = false;
  }
#endif

  return result;
}

// アイテムのレンジチェック(一括)
bool DX_CallbackValidNextMapRange (const void *pmem) {
  const TE097BCtrlTable *pNewControlTable = (TE097BCtrlTable *)pmem;
  bool pincfg = false;

  for (int i = 0; i < 12; i++) {
    if ((pNewControlTable->PinCfg[i] > tPinSPFUNC) || ((i != 9) && (pNewControlTable->PinCfg[i] == tPinAOUT))) {
      pincfg = true;
      break;
    } else {
      for (int j = 0; j < 12; j++) {
        if ((i != j) && (pNewControlTable->PinCfg[i] >= tPinPWM0) && (pNewControlTable->PinCfg[i] <= tPinMPC3) && (pNewControlTable->PinCfg[i] == pmemmap->PinCfg[j])) {
          pincfg = true;
          break;
        }
      }
    }
  }

  return (
#if DX_PROTOCOL_VER == 2
    ! ((pNewControlTable->Baudrate <= 4) || (pNewControlTable->Baudrate == 6))
    ||
#endif
    (pNewControlTable->OUT > 0xfff)
    || (pNewControlTable->OUT > 0xfff)
    || (pNewControlTable->IMU_OprMode > 12)
    || pincfg
  );
}

// NVM初期化時
void DX_CallbackInitNVMValue (void *pCtrl, const void *pNVM) {
  TE097BCtrlTable *pControlTable = (TE097BCtrlTable *)pCtrl;
  const TE097BCtrlTable *pNVMdata = (TE097BCtrlTable *)pNVM;

  update_PWMFreqency (0);

  pControlTable->PWMFrequency = pNVMdata->PWMFrequency;
  for (int i = 0; i < 6; i++) pControlTable->PWMDuty[i] = pNVMdata->PWMDuty[i];
  for (int i = 0; i < 12; i++) {
    pControlTable->PinCfg[i] = pNVMdata->PinCfg[i];
    switch (pControlTable->PinCfg[i]) {
      case tPinPWM0:    // 6
      case tPinPWM1:    // 7
      case tPinPWM2:    // 8
      case tPinPWM3:    // 9
      case tPinPWM4:    // 10
      case tPinPWM5:    // 11
        pin_config (i, (TMIOPinMode)pControlTable->PinCfg[i], pControlTable->PWMFrequency, pControlTable->PWMDuty[pControlTable->PinCfg[i] - tPinPWM0]);
        break;
      default:
        pin_config (i, (TMIOPinMode)pControlTable->PinCfg[i], pControlTable->PWMFrequency, 0);
        break;
    }
  }
  pControlTable->Terminator = pNVMdata->Terminator;
  Chip_GPIO_SetPinState (LPC_GPIO_PORT, 0, _TERM, (bool)pControlTable->Terminator);

  imu_opmode = (TOperationMode)(pControlTable->IMU_OprMode = pNVMdata->IMU_OprMode);
  imu_axisplace = (TAxisPlace)(pControlTable->IMU_AxisPlace = pNVMdata->IMU_AxisPlace);
}

// 読込パケット時
void DX_CallbackReadAction (void *pCtrl) {
  TE097BCtrlTable *pControlTable = (TE097BCtrlTable *)pCtrl;
  uint16_t w = 0;

  for (int i = 0; i < 12; i++) {
    pControlTable->ADV[i] = ADC_DR_RESULT (Chip_ADC_GetDataReg (LPC_ADC, i));
    w |= (Chip_GPIO_GetPinState (LPC_GPIO_PORT, 0, PinInfo[i].No) ? (1 << i) : 0);
  }
  pControlTable->IN = w;
  /*
  //  pControlTable->Capture[0] = pwd[0];
    uint32_t r = LPC_SCT->CAP[4].U >> 1;
    pControlTable->Capture[0] = r;
    pControlTable->Capture[1] = r>>16;
    pControlTable->Capture[2] = pwd[2];
    pControlTable->Capture[3] = pwd[3];
  */
}

// 書込パケット時
void DX_CallbackWriteAction (const void *pNewCtrl, void *pCtrl, bool force) {
  const TE097BCtrlTable *pNewControlTable = (TE097BCtrlTable *)pNewCtrl;
  TE097BCtrlTable *pControlTable = (TE097BCtrlTable *)pCtrl;
  bool update_sct = false;

  // LED On/Off
  if (pNewControlTable->LED != pControlTable->LED || force) {
    SetLED (pNewControlTable->LED);
  }
  // Pin Config
  for (int i = 0; i < 12; i++) {
    if (pNewControlTable->PinCfg[i] != pControlTable->PinCfg[i] || force) {
      switch (pNewControlTable->PinCfg[i]) {
        case tPinPWM0:    // 6
        case tPinPWM1:    // 7
        case tPinPWM2:    // 8
        case tPinPWM3:    // 9
        case tPinPWM4:    // 10
        case tPinPWM5:    // 11
          pin_config (i, (TMIOPinMode)pNewControlTable->PinCfg[i], pNewControlTable->PWMFrequency, pNewControlTable->PWMDuty[pNewControlTable->PinCfg[i] - tPinPWM0]);
          break;
        default:
          pin_config (i, (TMIOPinMode)pNewControlTable->PinCfg[i], pNewControlTable->PWMFrequency, 0);
          break;
      }
    }
  }
  // Terminator
  if (pNewControlTable->Terminator != pControlTable->Terminator || force) {
    Chip_GPIO_SetPinState (LPC_GPIO_PORT, 0, _TERM, (bool)pNewControlTable->Terminator);
  }
  // PWM Frequency
  if (pNewControlTable->PWMFrequency != pControlTable->PWMFrequency || force) {
    update_PWMFreqency (pNewControlTable->PWMFrequency);
    update_sct = true;
  }
  // PWM Duty
  for (int i = 0; i < 6; i++) {
    if (pNewControlTable->PWMDuty[i] != pControlTable->PWMDuty[i] || force || update_sct) {
      update_PWMDuty (i, pNewControlTable->PWMDuty[i]);
    }
  }
  // OUT
  if (pNewControlTable->OUT != pControlTable->OUT || force) {
    pControlTable->OUT &= 0xfff;
    for (int i = 0; i < 12; i++) Chip_GPIO_SetPinState (LPC_GPIO_PORT, 0, PinInfo[i].No, pNewControlTable->OUT & (1 << i) ? true : false);
  }
  // DAV
  if (pNewControlTable->DAV[9] != pControlTable->DAV[9] || force) {
    Chip_DAC_UpdateValue (LPC_DAC0, pNewControlTable->DAV[9] % 1024);
  }
  // IMU OPR MODE
  if (pNewControlTable->IMU_OprMode != pControlTable->IMU_OprMode || force) {
    imu_opmode = (TOperationMode)pNewControlTable->IMU_OprMode;
  }
  // IMU Axis Place
  if (pNewControlTable->IMU_AxisPlace != pControlTable->IMU_AxisPlace || force) {
    imu_axisplace = (TAxisPlace)pNewControlTable->IMU_AxisPlace;
  }
}

// FreeRTOS Stack over flow hook
//-----------------------------
void vApplicationStackOverflowHook (TaskHandle_t pxTask, char *pcTaskName) {
  taskDISABLE_INTERRUPTS();
  while (1) {
    for (int i = 0; i < 1000; i++) {
      SetLED (true);
      Delay_us (1000 - i);
      SetLED (false);
      Delay_us (i);
      ToggleLED();
    }
    Delay_ms (50);
    Chip_WWDT_Feed (LPC_WWDT);
  }
}

// TASK
//-----------------------------
// IMUの動作モード設定と測定値取得
void TaskIMUComm (void *pvParameters) {
  // TaskDXCommでコントロールテーブルが初期化される事を期待して少し待機
  vTaskDelay (100);

  static uint8_t div = 0;
  TOperationMode old_opm = imu_opmode;
  TAxisPlace old_axisp = imu_axisplace;

  typedef struct {
    Tbno055 data;
    uint8_t stat;
    uint8_t err;
  } __attribute__ ((packed)) TIMU;

  // I2C初期化
  LPC_SYSCON->I2C0CLKSEL = SYSCON_FLEXCOMMCLKSELSRC_MAINCLK;
  IIC_Init (LPC_I2C0, IIC_Fm);
  // IMU初期化
  init_imu (imu_opmode, imu_axisplace);

  static TIMU imu;
  portTickType t = xTaskGetTickCount();
  while (1) {
    // IMUデータ取得
    imu_read_datavalue (&imu.data);
    if (div++ % 8 == 0) imu_get_tst_stat_err (NULL, &imu.stat, &imu.err);
    // コントロールテーブルへ反映
#if DX_PROTOCOL_VER == 1
    DX_WriteMem (_ADDR_IMU, (uint8_t *)&imu, sizeof (TIMU));
#elif DX_PROTOCOL_VER == 2
    DX2_WriteMem (_ADDR_IMU, (uint8_t *)&imu, sizeof (TIMU));
#endif
    // 動作モード更新
    if (imu_opmode != old_opm) {
      imu_set_opmode (imu_opmode, NULL);
      old_opm = imu_opmode;
    }
    // センサ配置更新
    if (imu_axisplace != old_axisp) {
      imu_set_axix_place (imu_axisplace);
      old_axisp = imu_axisplace;
    }

    vTaskDelayUntil (&t, 10);
  }
}

#ifdef _USE_MYTASK
void MyTask (void *pvParameters) {
  portTickType t;
  static int16_t p[6] = {4600, 6800, 9000, 6800, 6800, 6800};
  static int16_t d[6] = { 15, -15, 15, -15, 15, -15 };
  uint32_t tog = 0;

  vTaskDelay (100);
#if DX_PROTOCOL_VER == 1
  DX_ReadMem (_ADDR_PWMDuty, (uint8_t *)&p, 12);
#elif DX_PROTOCOL_VER == 2
  DX2_ReadMem (_ADDR_PWMDuty, (uint8_t *)&p, 12);
#endif

  t = xTaskGetTickCount();

  while (1) {
    for (int i = 0; i < 6; i++) {
      p[i] += d[i];
      if (p[i] > 11000) {
        p[i] = 11000;
        d[i] *= -1;
      } else if (p[i] < 4000) {
        p[i] = 4000;
        d[i] *= -1;
      }
    }
#if DX_PROTOCOL_VER == 1
    DX_WriteMem (_ADDR_PWMDuty, (uint8_t *)&p, 12);
#elif DX_PROTOCOL_VER == 2
    DX2_WriteMem (_ADDR_PWMDuty, (uint8_t *)&p, 12);
#endif

    if (tog++ % 10 == 0) ToggleLED();

    vTaskDelayUntil (&t, 10);
  }
  /*
    while (1) {
      if (p[0] != 32767) p[0] = 32767;
      else p[0] = 4096;
  #if DX_PROTOCOL_VER == 1
      DX_WriteMem (_ADDR_PWMDuty, (uint8_t *)&p, 12);
  #elif DX_PROTOCOL_VER == 2
      DX2_WriteMem (_ADDR_PWMDuty, (uint8_t *)&p, 12);
  #endif
      vTaskDelayUntil (&t, 1000);
    }
  */
}
#endif

void TaskDXComm (void *pvParameters) {
#if DX_PROTOCOL_VER == 1

  DX_ClientCommTask();
#elif DX_PROTOCOL_VER == 2
  DX2_ClientCommTask();
#endif
}

// main
//-----------------------------
void wdt_feed (void) {
  Chip_WWDT_Feed (LPC_WWDT);
}
void sys_reset (void) {
  NVIC_SystemReset();
}

#if DX_PROTOCOL_VER == 1
static TDXClient_ConfParam ConfParam = {
  { (void *) &DEFAULT_DATA, sizeof (TE097BCtrlTable), wdt_feed, sys_reset, 0xF800, 256 },
  { us_init, us_deinit, us_setbaudrate, usart0_dma_tx_buff, usart0_dma_rx_buff, usart0_dma_tx_purge, usart0_dma_rx_purge, usart0_dma_putsbd, usart0_dma_getc },
  { DX_CallbackValidPrevMapRange, DX_CallbackValidNextMapRange, DX_CallbackInitNVMValue, DX_CallbackReadAction, DX_CallbackWriteAction }
};
#elif DX_PROTOCOL_VER == 2
static TDX2Client_ConfParam ConfParam = {
  { (void *) &DEFAULT_DATA, sizeof (TE097BCtrlTable), true, NULL, 0, NULL, wdt_feed, sys_reset, 0xF800, 256},
  { us_init, us_deinit, us_setbaudrate, usart0_dma_tx_buff, usart0_dma_rx_buff, usart0_dma_tx_purge, usart0_dma_rx_purge, usart0_dma_putsbd, usart0_dma_getc },
  { DX_CallbackValidPrevMapRange, DX_CallbackValidNextMapRange, DX_CallbackInitNVMValue, DX_CallbackReadAction, DX_CallbackWriteAction },
};
#endif

int main (void) {
  // ADCのクロック源を選択 (ADC_Initでは触れていない)
  LPC_SYSCON->ADCCLKSEL = 0; // Use fro_clk as source for ADC async clock (fro)
  LPC_SYSCON->ADCCLKDIV = 1; // Enable clock, and divide-by-1 at this clock divider
  // ADC初期化 (200kHz 全12チャンネル活性化 バーストモード)
  ADC_Init (200000, 0xfff);
  // DAC初期化
  Chip_DAC0_Init();
  // WDT初期化
  init_wdt();
//  NVIC_EnableIRQ (SCT_IRQn);

#if DX_PROTOCOL_VER == 1
  if (DX_Config (&ConfParam)) {
    pmemmap = (TE097BCtrlTable *)DX_GetPCtrlTalbe();
#elif DX_PROTOCOL_VER == 2
  if (DX2_Config (&ConfParam)) {
    pmemmap = (TE097BCtrlTable *)DX2_GetPCtrlTalbe();
#endif
    xTaskCreate (TaskDXComm, NULL, 220, NULL, 1, NULL);
    xTaskCreate (TaskIMUComm, NULL, 120, NULL, 1, NULL);
#ifdef _USE_MYTASK
    xTaskCreate (MyTask, NULL, 210, NULL, 1, NULL);
#endif
    vTaskStartScheduler();
  }
  while (1) {
    Chip_WWDT_Feed (LPC_WWDT);
    ToggleLED();
    Delay_ms (200);
  }
}
