﻿/*
  USART

  半二重のRS-485(CN4)ないしTTL(CN3) I/F、もしくはCN1-6,CN1-7を介したシリ
  アル通信。
  USART用のライブラリはポーリング・割り込み・DMAの3種類用意されている。
  いずれも使用する側のコード上の表面的な差は無いが、ポーリング版は一生懸
  命USARTのレジスタを監視しながら通信処理を行い、割り込み版はその名の通
  りレジスタの変化を割り込みで捉えながら通信処理を行い、DMA版は送受信デー
  タのやりとりをDMAを介して行うといった違いがある。

  ・端子の初期化やマクロは e097b_def.h と e097b_usart.h を参照
  ・物理的な送受信口にDXMIOのCN3/CN4を用いる場合は USART_TYPE に1～3、
    CN1-6/CN1-7の場合は4～6を定義した後に e097b_def.h をincludeする事
  ・e097b_usart.h に定義されたマクロを用いてUSARTの関数名が置換
     us_init        初期化
     us_setbaudrate ボーレート変更
     us_tx_buff     残り送信バイト数
     us_rx_buff     受信データの未取り出しバイト数
     us_tx_purge    送信データ一掃
     us_rx_purge    受信データ一掃
     us_putc        1バイト送信
     us_puts        文字列送信
     us_putsb       バイト列送信
     us_getc        1バイト受信待ち
     us_gets        文字列受信待ち
     us_printf      書式化文字列送信
     us_scanf       書式化文字列受信
  ・主にポーリング版は割り込みルーチン内で使い、割り込み版は省電力モー
    ドからの復帰に使い、DMA版は送受データとUSART間のやりとり時にCPUの能
    力をほとんど使わないのでその間に他の処理を行うといった使い分けができ
    る
  ・書式化文字列関数(scanf,printf)は簡易版なのであまり期待しないこと
  ・いずれも初期化時に2バイト以上の送受信バッファを必要とする
  ・ポーリング版のバッファは他との体裁を合わせるためだけに引数を受け取る
    ので、広大なバッファを用意する意味は皆無。
  ・割り込み及びDMA版の受信バッファサイズは大きいほどバッファからデータ
    を取り出すタイミングに余裕ができる
  ・USARTのクロック源は e097_def.h の初期化部で32MHzを供給する設定を行っ
    ているので、最大ボーレートはそれを8で割った4Mbpsとなる
  ・USART3～4にはRTS端子が何故か無いので、本ボードの半二重I/F制御用には
    USART0～2のみが使用できる

  ・CN3とCN4はソフト的には区別されず、送信は両者同時、受信は論理和で処理
  ・CN3とCN4の半二重処理はトランシーバ部分のみで、それ以外は実質全二重
  ・半二重の場合はUSARTの送信開始と共にRTSが自動的にHIGHになりトランシー
    バーの送信がイネーブルされ、データを完全に送信し終えるとRTSがLOWにな
    りトランシーバーの受信がイネーブルになる
  ・サンプルコード上では1バイト受信毎に返信しているが、半二重か全二重を
    全く意識していないので、連続的に受信が発生するとCN3/CN4の場合は送信
    データとコンフリクトが生じてターミナル上ではまともな表示が得られない
  ・RS-485トランシーバに内蔵されたターミネータはプログラムからON/OFFでき、
    デフォルトはPIO_Configureによる端子の初期化時にON
  ・TTL I/Fのみを使用する際は使用しないRS-485に装備されたターミネータを
    必ずONにし受信回路のハイインピーダンス状態を回避する事
  ・CN1-6/CN1-7を使用する場合は完全な全二重
 */

#include  <chip.h>
// ボードの端子割り当てと最小限の初期化
#include  "e097b_def.h"

// USART_TYPEマクロで使用するUSARTと機能を選択
// CN3/CN4     1:ポーリング, 2:割り込み, 3:DMA
// CN1-6,CN1-7 4:ポーリング, 5:割り込み, 6:DMA
#ifndef USART_TYPE
  #define USART_TYPE 1
#endif
#include  "e097b_usart.h"

#include  <wl_delay.h>

// main
//-----------------------------
void main (void) {
  // シリアル通信バッファ
  uint8_t txd[20], rxd[20];

  // USARTを初期化
  // USARTのクロックソースはfrg0_src_clkを元にしたFRG0を選択しているので、
  // Chip_Clock_GetFRGClockRate (0)は32,000,000Hzを返す
  us_init (Chip_Clock_GetFRGClockRate (0), 115200, USIP_8N1, txd, sizeof (txd), rxd, sizeof (rxd));

  // 文字列送信
  us_puts ("\nHELLO!\n");

  char s[100], c;
  int i = 0;
  for (;;) {
    // 1文字受信 (受信データが得られるまで待機)
    c = us_getc ();
    // 1文字送信 (受信したデータをそのまま返信)
    us_putc (c);
    // 書式化文字列送信
    us_printf ("(%d)", i);
    switch (c) {
      case '0':
        us_puts ("\n>Input string = ");
        // 書式化文字列受信
        us_scanf ("%s", &s);
        us_printf ("\n string = %s\n", s);
        break;
      case '1':
        us_puts ("\n>Input number = ");
        us_scanf ("%d", &i);
        us_printf ("\n number = %d\n", i);
        break;
      case '2':
        us_puts ("\n>Term ON\n");
        SetTerminator (true);
        break;
      case '3':
        us_puts ("\n>Term OFF\n");
        SetTerminator (false);
        break;
      case '!':
        us_puts ("\n>BYE!\n ");
        Delay_ms (50);
        NVIC_SystemReset();
        break;
    }
    i++;
  }
}
