﻿/*
  ICSクライアント

  KRSシリーズと同じネットワークに配置してクライアントとして機能させる。

  ・GPIOを個別に任意のモードに設定可能
  ・IMUを常時モニタしコントロールテーブルへ逐次反映
  ・キャプチャ及びカウンタ機能はペンディング
  ・汎用コマンドに応答するのが基本だが、便宜上他のコマンドにもダミー応
    答する
 */

#include  <FreeRTOS.h>
#include  <task.h>

#include  <chip.h>
// ボードの端子割り当てと最小限の初期化
#include  "e097b_def.h"
// IMU用サブルーチン
#include  "e097b_imu.h"
#include  <us.h>
#include  <icsclient.h>
#include  <wl_delay.h>
#include  <wl_adc.h>
#include  <wl_i2c.h>

// マクロと変数
//-----------------------------
// DXMIO用Pin Mode
typedef enum { tPinDIN = 0, tPinDIN_PU, tPinDIN_PD, tPinDOUT, tPinDOUT_OC, tPinAIN, tPinPWM0, tPinPWM1, tPinPWM2, tPinPWM3, tPinPWM4, tPinPWM5, tPinMPW0, tPinMPW1, tPinMPW2, tPinMPW3, tPinMPC0, tPinMPC1, tPinMPC2, tPinMPC3, tPinAOUT, tPinSPFUNC } TMIOPinMode;

// Pin information
typedef struct {
  uint8_t               No;
  CHIP_PINx_T           CPin;
  CHIP_SWM_PIN_FIXED_T  AD;
} TPinInfo;

// 外部端子
const TPinInfo PinInfo[12] = {
  { _GPIO0,  IOCON_PIO0_7,  SWM_FIXED_ADC_0  },
  { _GPIO1,  IOCON_PIO0_6,  SWM_FIXED_ADC_1  },
  { _GPIO2,  IOCON_PIO0_14, SWM_FIXED_ADC_2  },
  { _GPIO3,  IOCON_PIO0_23, SWM_FIXED_ADC_3  },
  { _GPIO4,  IOCON_PIO0_22, SWM_FIXED_ADC_4  },
  { _GPIO5,  IOCON_PIO0_21, SWM_FIXED_ADC_5  },
  { _GPIO6,  IOCON_PIO0_20, SWM_FIXED_ADC_6  },
  { _GPIO7,  IOCON_PIO0_19, SWM_FIXED_ADC_7  },
  { _GPIO8,  IOCON_PIO0_18, SWM_FIXED_ADC_8  },
  { _GPIO9,  IOCON_PIO0_17, SWM_FIXED_ADC_9  },
  { _GPIO10, IOCON_PIO0_13, SWM_FIXED_ADC_10 },
  { _GPIO11, IOCON_PIO0_4,  SWM_FIXED_ADC_11 }
};

TMIOPinMode previous_setting[12] = { tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN, tPinDIN };

TOperationMode imu_opmode = opmode_NDOF;
TAxisPlace imu_axisplace = axisplace_P0;

// シリアル通信バッファ
static uint8_t txd[300], rxd[300];

// コントロールテーブル
typedef struct {
  uint16_t  ModelNumber;      //
  uint8_t   FirmwareVersion;  //
  uint8_t   ID;               // ID (R/W & NVM)
  uint8_t   Baudrate;         // ボーレート (R/W & NVM)
  uint8_t   WriteNVM;         // 不揮発データ書き込みフラグ (R/W)
  // これ以後がユーザ定義アイテム
  uint8_t   LED;              // LED2  (R/W)
  uint8_t   Terminator;       // 終端抵抗 (R/W & NVM)
  uint8_t   PinCfg[12];       // GPIO端子機能設定 (R/W & NVM)
  uint16_t  PWMFrequency;     // PWM周波数設定 (R/W & NVM)
  uint16_t  PWMDuty[6];       // PWMデューティー設定 (R/W & NVM)
  uint16_t  Capture[4];       // パルス設定・計測値 (R/W)
  uint16_t  OUT;              // GPIO出力設定値 (R/W)
  uint16_t  IN;               // GPIO入力値 (R)
  uint16_t  ADV[12];          // アナログ電圧測定値 (R)
  uint16_t  DAV;              // アナログ電圧設定値 (R/W)
  uint8_t   IMU_OprMode;      // IMU動作モード (R/W & NVM)
  uint8_t   IMU_AxisPlace;    // IMUセンサ配置 (R/W & NVM)
  Tbno055   IMU;              // IMU測定値 (R)
  uint8_t   IMU_stat;         // IMUステータス値 (R)
  uint8_t   IMU_err;          // IMUエラー値 (R)
} __attribute__ ((packed)) TE097BCtrlTable;

// コントロールテーブルのポインタ
TE097BCtrlTable  *pmemmap;

// コントロールテーブルの各アイテムのオフセットアドレス
#define _ADDR_PinCfg        ((int)&((TE097BCtrlTable*)0)->PinCfg)
#define _ADDR_PWMFrequency  ((int)&((TE097BCtrlTable*)0)->PWMFrequency)
#define _ADDR_PWMDuty       ((int)&((TE097BCtrlTable*)0)->PWMDuty)
#define _ADDR_Capture       ((int)&((TE097BCtrlTable*)0)->Capture)
#define _ADDR_OUT           ((int)&((TE097BCtrlTable*)0)->OUT)
#define _ADDR_IN            ((int)&((TE097BCtrlTable*)0)->IN)
#define _ADDR_ADV           ((int)&((TE097BCtrlTable*)0)->ADV)
#define _ADDR_DAV           ((int)&((TE097BCtrlTable*)0)->DAV)
#define _ADDR_IMU_OprMode   ((int)&((TE097BCtrlTable*)0)->IMU_OprMode)
#define _ADDR_IMU           ((int)&((TE097BCtrlTable*)0)->IMU)
#define _ADDR_IMU_stat      ((int)&((TE097BCtrlTable*)0)->IMU_stat)
#define _ADDR_IMU_err       ((int)&((TE097BCtrlTable*)0)->IMU_err)

// コントロールテーブルデフォルト値
static const TE097BCtrlTable DEFAULT_DATA = {
  ModelNumber     : 0x4023,
  FirmwareVersion : 'A',
  ID              : 1,
  Baudrate        : 0,
  WriteNVM        : 0,
  LED             : 0,
  Terminator      : 1,

  IMU_OprMode     : opmode_NDOF,
  IMU_AxisPlace   : axisplace_P0,
};

// PWM
//-----------------------------
static void init_SCT (void) {
  static bool inited = false;

  if (!inited) {
    Chip_SCT_Init (LPC_SCT);
    inited = true;
  }
}

static void update_PWMFreqency (uint16_t freq) {
  static uint16_t oldfreq = 0;

  if (freq == 0) {
    Chip_SCTPWM_Stop (LPC_SCT);
    Chip_SCT_DeInit (LPC_SCT);
  } else {
    init_SCT();
    if (oldfreq != freq) Chip_SCTPWM_SetRate (LPC_SCT, freq);
  }
  oldfreq = freq;
}

static void update_PWMDuty (uint8_t ch, uint16_t duty) {
  uint32_t d;

  Chip_SCTPWM_SetOutPin (LPC_SCT, ch + 1, ch);
  if (duty == 0)          d = 0;
  else if (duty == 65535) d = Chip_SCTPWM_GetTicksPerCycle (LPC_SCT) + 1;
  else                    d = ((uint64_t)duty * (uint64_t)Chip_SCTPWM_GetTicksPerCycle (LPC_SCT)) / (uint64_t) (65535UL);
  Chip_SCTPWM_SetDutyCycle (LPC_SCT, ch + 1, d);
  Chip_SCTPWM_Start (LPC_SCT);
}

// pin configuration
//-----------------------------
static void pin_config (uint8_t adch, TMIOPinMode pm, uint16_t freq, uint16_t duty) {
  static const uint8_t SCT_OUT_pin_movable[6] = { SWM_SCTOUT_0_O, SWM_SCTOUT_1_O, SWM_SCTOUT_2_O, SWM_SCTOUT_3_O, SWM_SCTOUT_4_O, SWM_SCTOUT_5_O };
  static const uint8_t SCT_IN_pin_movable[4] = { SWM_SCTIN_0_I, SWM_SCTIN_1_I, SWM_SCTIN_2_I, SWM_SCTIN_3_I };
  TPin pin = { PinInfo[adch].No, PIO_TYPE_FIXED, 0, PIO_MODE_DEFAULT };
  uint8_t ch;

  // 割り当て済みSCTの切断
  if ((previous_setting[adch] >= tPinPWM0) && (previous_setting[adch] <= tPinPWM5)) {
    ch = previous_setting[adch] - tPinPWM0;
    Chip_Clock_EnablePeriphClock (SYSCON_CLOCK_SWM);
    Chip_SWM_MovablePinAssign (SCT_OUT_pin_movable[ch], 0xff);
    Chip_Clock_DisablePeriphClock (SYSCON_CLOCK_SWM);
  }
  if ((previous_setting[adch] >= tPinMPW0) && (previous_setting[adch] <= tPinMPW3)) {
    ch = previous_setting[adch] - tPinMPW0;
    Chip_Clock_EnablePeriphClock (SYSCON_CLOCK_SWM);
    Chip_SWM_MovablePinAssign (SCT_IN_pin_movable[ch], 0xff);
    Chip_Clock_DisablePeriphClock (SYSCON_CLOCK_SWM);
  }
  if ((previous_setting[adch] >= tPinMPC0) && (previous_setting[adch] <= tPinMPC3)) {
    ch = previous_setting[adch] - tPinMPC0;
    Chip_Clock_EnablePeriphClock (SYSCON_CLOCK_SWM);
    Chip_SWM_MovablePinAssign (SCT_IN_pin_movable[ch], 0xff);
    Chip_Clock_DisablePeriphClock (SYSCON_CLOCK_SWM);
  }

  pin.gpio_no = PinInfo[adch].No;

  if (pin.gpio_no == _GPIO9) LPC_IOCON->PIO0_17 &= ~ (1 << IOCON_PIO_DACMODE_P);

  switch (pm) {
    case tPinDIN:     // 0
      PIO_Configure (&pin, 1);
      pin.pin_type = PIO_TYPE_INPUT;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinDIN_PU:  // 1
      pin.pin_type = PIO_TYPE_INPUT;
      pin.pin_mode = PIO_MODE_PULLUP;
      break;
    case tPinDIN_PD:  // 2
      pin.pin_type = PIO_TYPE_INPUT;
      pin.pin_mode = PIO_MODE_PULLDOWN;
      break;
    case tPinDOUT:    // 3
      pin.pin_type = PIO_TYPE_OUTPUT;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinDOUT_OC: // 4
      pin.pin_type = PIO_TYPE_OUTPUT;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinAIN:     // 5
      pin.pin_type = PIO_TYPE_FIXED;
      pin.pin_movable_fixed = PinInfo[adch].AD;
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinPWM0:    // 6
    case tPinPWM1:    // 7
    case tPinPWM2:    // 8
    case tPinPWM3:    // 9
    case tPinPWM4:    // 10
    case tPinPWM5:    // 11
      ch = pm - tPinPWM0;
      // disconnect
      Chip_Clock_EnablePeriphClock (SYSCON_CLOCK_SWM);
      Chip_SWM_MovablePinAssign (SCT_OUT_pin_movable[ch], 0xff);
      Chip_Clock_DisablePeriphClock (SYSCON_CLOCK_SWM);

      pin.pin_type = PIO_TYPE_MOVABLE;
      pin.pin_movable_fixed = SCT_OUT_pin_movable[ch];
      pin.pin_mode = PIO_MODE_DEFAULT;
      update_PWMFreqency (freq);
      update_PWMDuty (ch, duty);
      break;
    case tPinMPW0:    // 12
    case tPinMPW1:    // 13
    case tPinMPW2:    // 14
    case tPinMPW3:    // 15
      ch = pm - tPinMPW0;
      // disconnect
      Chip_Clock_EnablePeriphClock (SYSCON_CLOCK_SWM);
      Chip_SWM_MovablePinAssign (SCT_IN_pin_movable[ch], 0xff);
      Chip_Clock_DisablePeriphClock (SYSCON_CLOCK_SWM);

      pin.pin_type = PIO_TYPE_MOVABLE;
      pin.pin_movable_fixed = SCT_IN_pin_movable[ch];
      pin.pin_mode = PIO_MODE_DEFAULT;
//      init_SCTCapture ();
      break;
    case tPinMPC0:    // 16
    case tPinMPC1:    // 17
    case tPinMPC2:    // 18
    case tPinMPC3:    // 19
      ch = pm - tPinMPC0;
      // disconnect
      Chip_Clock_EnablePeriphClock (SYSCON_CLOCK_SWM);
      Chip_SWM_MovablePinAssign (SCT_IN_pin_movable[ch], 0xff);
      Chip_Clock_DisablePeriphClock (SYSCON_CLOCK_SWM);

      pin.pin_type = PIO_TYPE_MOVABLE;
      pin.pin_movable_fixed = SCT_IN_pin_movable[ch];
      pin.pin_mode = PIO_MODE_DEFAULT;
      break;
    case tPinAOUT:    // 20
      // only pio0_17
      pin.pin_type = PIO_TYPE_FIXED;
      pin.pin_movable_fixed = SWM_FIXED_DAC_0;
      pin.pin_mode = PIO_MODE_DEFAULT;
      LPC_IOCON->PIO0_17 |= (1 << IOCON_PIO_DACMODE_P);
    case tPinSPFUNC:  // 21
      break;
  }
  previous_setting[adch] = pm;
  PIO_Configure (&pin, 1);
}

// WDT
//-----------------------------
static void init_wdt (void) {
  uint32_t wdtFreq;

  Chip_Clock_SetWDTOSC (WDTLFO_OSC_0_60, 64);     // Freq = 0.6Mhz, divided by 64. WDT_OSC should be 9.375khz
  Chip_SYSCON_PowerUp (SYSCON_PDRUNCFG_WDTOSC);   // Enable the power to the WDT
  wdtFreq = Chip_Clock_GetWDTOSCRate() / 4;       // The WDT divides the input frequency into it by 4
  Chip_WWDT_Init (LPC_WWDT);                      // Initialize WWDT (also enables WWDT clock)
  /* Set watchdog feed time constant to approximately 2s
     Set watchdog warning time to 512 ticks after feed time constant
     Set watchdog window time to 3s */
  Chip_WWDT_SetTimeOut (LPC_WWDT, wdtFreq * 2);
  Chip_WWDT_SetWarning (LPC_WWDT, 512);
  Chip_WWDT_SetWindow (LPC_WWDT, wdtFreq * 3);
  Chip_WWDT_SetOption (LPC_WWDT, WWDT_WDMOD_WDRESET);  // Configure WWDT to reset on timeout
  Chip_WWDT_ClearStatusFlag (LPC_WWDT, WWDT_WDMOD_WDTOF | WWDT_WDMOD_WDINT); // Clear watchdog warning and timeout interrupts
  Chip_WWDT_Start (LPC_WWDT);  // Start watchdog
}

// USART
//-----------------------------
static const uint32_t baud[3] = { 115200, 625000, 1250000 };

// 初期化
static uint32_t us_init (uint8_t ind) {
  if (ind > 2) ind = 2;
  return usart0_dma_init (Chip_Clock_GetFRGClockRate (0), baud[ind], USIP_8E1, txd, sizeof (txd), rxd, sizeof (rxd));
}

// ボーレート更新
static uint32_t us_setbaudrate (uint8_t ind) {
  if (ind > 2) ind = 2;
  return usart0_dma_setbaudrate (Chip_Clock_GetFRGClockRate (0), baud[ind]);
}

// 終了
static void us_deinit (void) { }

// ICSクライアントコールバック
//-----------------------------
// アイテムのレンジチェック(個別)
static bool ICS_CallbackValidPrevMapRange (const void *pmem, uint8_t u_st, uint8_t u_len) {
  bool result = true;
  const TE097BCtrlTable *pNewControlTable = pmem;

  for (int i = 0; i < 12; i++) {
    if (IS_SAFE_VAR (pNewControlTable, pNewControlTable->PinCfg[i], u_st, u_len)) {
      if ((pNewControlTable->PinCfg[i] > tPinSPFUNC) || ((i != 9) && (pNewControlTable->PinCfg[i] == tPinAOUT))) {
        result = false;
        break;
      } else {
        for (int j = 0; j < 12; j++) {
          if ((i != j) && (pNewControlTable->PinCfg[i] >= tPinPWM0) && (pNewControlTable->PinCfg[i] <= tPinMPC3) && (pNewControlTable->PinCfg[i] == pmemmap->PinCfg[j])) {
            result = false;
            break;
          }
        }
      }
    }
  }
  if (IS_SAFE_VAR (pNewControlTable, pNewControlTable->PWMFrequency, u_st, u_len)) {
    if (pNewControlTable->PWMFrequency > 50000) result = false;
  }
  if (IS_SAFE_VAR (pNewControlTable, pNewControlTable->OUT, u_st, u_len)) {
    if (pNewControlTable->OUT > 0xfff) result = false;
  }
  if (IS_SAFE_VAR (pNewControlTable, pNewControlTable->OUT, u_st, u_len)) {
    if (pNewControlTable->OUT > 0xfff) result = false;
  }
  if (IS_SAFE_VAR (pNewControlTable, pNewControlTable->IMU_OprMode, u_st, u_len)) {
    if (pNewControlTable->IMU_OprMode > 12) result = false;
  }
  if (IS_SAFE_VAR (pNewControlTable, pNewControlTable->Baudrate, u_st, u_len)) {
    if (! ((pNewControlTable->Baudrate <= 4) || (pNewControlTable->Baudrate == 6))) result = false;
  }
  return result;
}

// アイテムのレンジチェック(一括)
bool ICS_CallbackValidNextMapRange (const void *pmem) {
  const TE097BCtrlTable *pNewControlTable = pmem;
  bool pincfg = false;

  for (int i = 0; i < 12; i++) {
    if ((pNewControlTable->PinCfg[i] > tPinSPFUNC) || ((i != 9) && (pNewControlTable->PinCfg[i] == tPinAOUT))) {
      pincfg = true;
      break;
    } else {
      for (int j = 0; j < 12; j++) {
        if ((i != j) && (pNewControlTable->PinCfg[i] >= tPinPWM0) && (pNewControlTable->PinCfg[i] <= tPinMPC3) && (pNewControlTable->PinCfg[i] == pmemmap->PinCfg[j])) {
          pincfg = true;
          break;
        }
      }
    }
  }

  return (
    ! ((pNewControlTable->Baudrate <= 4) || (pNewControlTable->Baudrate == 6))
    || (pNewControlTable->OUT > 0xfff)
    || (pNewControlTable->OUT > 0xfff)
    || (pNewControlTable->IMU_OprMode > 12)
    || pincfg
  );
}

// NVM初期化時
void ICS_CallbackInitNVMValue (void *pCtrl, const void *pNVM) {
  TE097BCtrlTable *pControlTable = pCtrl;
  const TE097BCtrlTable *pNVMdata = pNVM;

  update_PWMFreqency (0);

  pControlTable->PWMFrequency = pNVMdata->PWMFrequency;
  for (int i = 0; i < 6; i++) pControlTable->PWMDuty[i] = pNVMdata->PWMDuty[i];
  for (int i = 0; i < 12; i++) {
    pControlTable->PinCfg[i] = pNVMdata->PinCfg[i];
    switch (pControlTable->PinCfg[i]) {
      case tPinPWM0:    // 6
      case tPinPWM1:    // 7
      case tPinPWM2:    // 8
      case tPinPWM3:    // 9
      case tPinPWM4:    // 10
      case tPinPWM5:    // 11
        pin_config (i, pControlTable->PinCfg[i], pControlTable->PWMFrequency, pControlTable->PWMDuty[pControlTable->PinCfg[i] - tPinPWM0]);
        break;
      default:
        pin_config (i, pControlTable->PinCfg[i], pControlTable->PWMFrequency, 0);
        break;
    }
  }
  pControlTable->Terminator = pNVMdata->Terminator;
  Chip_GPIO_SetPinState (LPC_GPIO_PORT, 0, _TERM, (bool)pControlTable->Terminator);

  imu_opmode = pControlTable->IMU_OprMode = pNVMdata->IMU_OprMode;
  imu_axisplace = pControlTable->IMU_AxisPlace = pNVMdata->IMU_AxisPlace;
}

// 読込パケット時
void ICS_CallbackReadAction (void *pCtrl) {
  TE097BCtrlTable *pControlTable = pCtrl;
  uint16_t w = 0;

  for (int i = 0; i < 12; i++) {
    pControlTable->ADV[i] = ADC_DR_RESULT (Chip_ADC_GetDataReg (LPC_ADC, i));
    w |= (Chip_GPIO_GetPinState (LPC_GPIO_PORT, 0, PinInfo[i].No) ? (1 << i) : 0);
  }
  pControlTable->IN = w;
  /*
  //  pControlTable->Capture[0] = pwd[0];
    uint32_t r = LPC_SCT->CAP[4].U >> 1;
    pControlTable->Capture[0] = r;
    pControlTable->Capture[1] = r>>16;
    pControlTable->Capture[2] = pwd[2];
    pControlTable->Capture[3] = pwd[3];
  */
}

// 書込パケット時
void ICS_CallbackWriteAction (const void *pNewCtrl, void *pCtrl, bool force) {
  const TE097BCtrlTable *pNewControlTable = pNewCtrl;
  TE097BCtrlTable *pControlTable = pCtrl;
  bool update_sct = false;

  // LED On/Off
  if (pNewControlTable->LED != pControlTable->LED || force) {
    SetLED (pNewControlTable->LED);
  }
  // Pin Config
  for (int i = 0; i < 12; i++) {
    if (pNewControlTable->PinCfg[i] != pControlTable->PinCfg[i] || force) {
      switch (pNewControlTable->PinCfg[i]) {
        case tPinPWM0:    // 6
        case tPinPWM1:    // 7
        case tPinPWM2:    // 8
        case tPinPWM3:    // 9
        case tPinPWM4:    // 10
        case tPinPWM5:    // 11
          pin_config (i, pNewControlTable->PinCfg[i], pNewControlTable->PWMFrequency, pNewControlTable->PWMDuty[pNewControlTable->PinCfg[i] - tPinPWM0]);
          break;
        default:
          pin_config (i, pNewControlTable->PinCfg[i], pNewControlTable->PWMFrequency, 0);
          break;
      }
    }
  }
  // Terminator
  if (pNewControlTable->Terminator != pControlTable->Terminator || force) {
    Chip_GPIO_SetPinState (LPC_GPIO_PORT, 0, _TERM, (bool)pNewControlTable->Terminator);
  }
  // PWM Frequency
  if (pNewControlTable->PWMFrequency != pControlTable->PWMFrequency || force) {
    update_PWMFreqency (pNewControlTable->PWMFrequency);
    update_sct = true;
  }
  // PWM Duty
  for (int i = 0; i < 6; i++) {
    if (pNewControlTable->PWMDuty[i] != pControlTable->PWMDuty[i] || force || update_sct) {
      update_PWMDuty (i, pNewControlTable->PWMDuty[i]);
    }
  }
  // OUT
  if (pNewControlTable->OUT != pControlTable->OUT || force) {
    pControlTable->OUT &= 0xfff;
    for (int i = 0; i < 12; i++) Chip_GPIO_SetPinState (LPC_GPIO_PORT, 0, PinInfo[i].No, pNewControlTable->OUT & (1 << i) ? true : false);
  }
  // DAV
  if (pNewControlTable->DAV != pControlTable->DAV || force) {
    Chip_DAC_UpdateValue (LPC_DAC0, pNewControlTable->DAV % 1024);
  }
  // IMU OPR MODE
  if (pNewControlTable->IMU_OprMode != pControlTable->IMU_OprMode || force) {
    imu_opmode = pNewControlTable->IMU_OprMode;
  }
  // IMU Axis Place
  if (pNewControlTable->IMU_AxisPlace != pControlTable->IMU_AxisPlace || force) {
    imu_axisplace = pNewControlTable->IMU_AxisPlace;
  }
}

// FreeRTOS Stack over flow hook
//-----------------------------
void vApplicationStackOverflowHook (TaskHandle_t pxTask, char *pcTaskName) {
  taskDISABLE_INTERRUPTS();
  while (1) {
    for (int i = 0; i < 1000; i++) {
      SetLED (true);
      Delay_us (1000 - i);
      SetLED (false);
      Delay_us (i);
      ToggleLED();
    }
    Delay_ms (50);
    Chip_WWDT_Feed (LPC_WWDT);
  }
}

// TASK
//-----------------------------
// IMUの動作モード設定と測定値取得
void TaskIMUComm (void *pvParameters) {
  // TaskICSCommでコントロールテーブルが初期化される事を期待して少し待機
  vTaskDelay (100);

  static uint8_t div = 0;
  TOperationMode old_opm = imu_opmode;
  TAxisPlace old_axisp = imu_axisplace;

  typedef struct {
    Tbno055 data;
    uint8_t stat;
    uint8_t err;
  } __attribute__ ((packed)) TIMU;

  // I2C初期化
  LPC_SYSCON->I2C0CLKSEL = SYSCON_FLEXCOMMCLKSELSRC_MAINCLK;
  IIC_Init (LPC_I2C0, IIC_Fm);
  // IMU初期化
  init_imu (imu_opmode, imu_axisplace);

  static TIMU imu;
  portTickType t = xTaskGetTickCount();
  while (1) {
    // IMUデータ取得
    imu_read_datavalue (&imu.data);
    if (div++ % 8 == 0) imu_get_tst_stat_err (NULL, &imu.stat, &imu.err);
    // コントロールテーブルへ反映
    ICS_WriteMem (_ADDR_IMU, (uint8_t *)&imu, sizeof (TIMU));
    // 動作モード更新
    if (imu_opmode != old_opm) {
      imu_set_opmode (imu_opmode, NULL);
      old_opm = imu_opmode;
    }
    // センサ配置更新
    if (imu_axisplace != old_axisp) {
      imu_set_axix_place (imu_axisplace);
      old_axisp = imu_axisplace;
    }

    vTaskDelayUntil (&t, 10);
  }
}

void TaskICSComm (void *pvParameters) {
  ICS_ClientCommTask();
}

// main
//-----------------------------
void wdt_feed (void) {
  Chip_WWDT_Feed (LPC_WWDT);
}
void sys_reset (void) {
  NVIC_SystemReset();
}

static TICSClient_ConfParam ConfParam = {
  { (void *) &DEFAULT_DATA, sizeof (TE097BCtrlTable), wdt_feed, sys_reset, 0xF800, 256 },
  { us_init, us_deinit, us_setbaudrate, usart0_dma_tx_buff, usart0_dma_rx_buff, usart0_dma_tx_purge, usart0_dma_rx_purge, usart0_dma_putsbd, usart0_dma_getc },
  { ICS_CallbackValidPrevMapRange, ICS_CallbackValidNextMapRange, ICS_CallbackInitNVMValue, ICS_CallbackReadAction, ICS_CallbackWriteAction }
};

void main (void) {
  // ICSのボーレートに合わせて
  // frg0clk = 60,000,000 / (1 + FRG0MUL / (FRG0DIV + 1)) = 60,000,000 / 2 = 30MHz
  LPC_SYSCON->FRG0MULT = 1;
  LPC_SYSCON->FRG0DIV = 0;

  // ADCのクロック源を選択
  LPC_SYSCON->ADCCLKSEL = 0; // Use fro_clk as source for ADC async clock (fro)
  LPC_SYSCON->ADCCLKDIV = 1; // Enable clock, and divide-by-1 at this clock divider
  // ADC初期化 (200kHz 全12チャンネル活性化 バーストモード)
  ADC_Init (200000, 0xfff);
  // DAC初期化
  Chip_DAC0_Init();
  // WDT初期化
  init_wdt();
//  NVIC_EnableIRQ (SCT_IRQn);

  if (ICS_Config (&ConfParam)) {
    pmemmap = ICS_GetPCtrlTalbe();
    xTaskCreate (TaskICSComm, NULL, 210, NULL, 1, NULL);
    xTaskCreate (TaskIMUComm, NULL, 120, NULL, 1, NULL);
    vTaskStartScheduler();
  }
  while (1) {
    Chip_WWDT_Feed (LPC_WWDT);
    ToggleLED();
    Delay_ms (200);
  }
}
