﻿/*
  I2C (BME680)

  コードの大半はbme680.c任せなので、I2Cらしいところは i2c_write・
  i2c_read 関数程度。

  ・GPIO2をSCL1、GPIO3をSDA1に割り当て
  ・BME680は5V電源で使用できる秋月電子社のモジュールを使用
    https://akizukidenshi.com/catalog/g/gK-14469/
     GPIO2 <-> SCL
     GPIO3 <-> SDA
  ・I2Cバスのプルアップはモジュール側で処置
  ・ドライバはBosh社提供のものを使用
    https://github.com/BoschSensortec/BME680_driver
 */

#include  <us.h>
#include  <chip.h>
// ボードの端子割り当てと最小限の初期化
#include  "e097b_def.h"
#ifndef USART_TYPE
  #define USART_TYPE 1
#endif
#include  "e097b_usart.h"
// I2C用関数
#include  <wl_i2c.h>
#include  <wl_delay.h>

#include  "bme680.h"
#include  "bme680.c" // 横着してライブラリソースをそのままインクルード

// マクロ・グローバル変数など
//-----------------------------
#define cls()           us_puts ("\x1b[2J");
#define gotoxy(xx, yy)  us_printf ("\x1b[%d;%dH", yy + 1, xx + 1);

//#define _USE_GAS_SENSOR

// I2Cフックルーチン
//-----------------------------
// I2Cのパラメータ
static const TIICDataSpec ds2 = { IIC_Reg8, IIC_Data8, IIC_LittleEndian };

// BME680のライブラリが要求するI2Cアクセス関数のフォーマットに変換
int8_t i2c_write (uint8_t dev_id, uint8_t reg_addr, uint8_t *reg_data, uint16_t len) {
  uint8_t a = reg_addr;
  return IIC_Write (LPC_I2C1, dev_id, &a, reg_data, len, &ds2, NULL) ? 0 : 1;
}

int8_t i2c_read (uint8_t dev_id, uint8_t reg_addr, uint8_t *reg_data, uint16_t len) {
  uint8_t a = reg_addr;
  return IIC_Read (LPC_I2C1, dev_id, &a, reg_data, len, &ds2, NULL) ? 0 : 1;
}

// main
//-----------------------------
int main (void) {
  const TPin pins[] = {
    { _GPIO2, PIO_TYPE_MOVABLE,  SWM_I2C1_SCL_IO, PIO_MODE_OPENDRAIN }, // I2C1 SCL
    { _GPIO3, PIO_TYPE_MOVABLE,  SWM_I2C1_SDA_IO, PIO_MODE_OPENDRAIN }, // I2C1 SDA
  };
  static uint8_t txb[100], rxb[10];

  // GPIO初期化
  PIO_Configure (pins, PIO_LISTSIZE (pins));

  // USART初期化
  us_init (Chip_Clock_GetFRGClockRate (0), 115200, USIP_8N1, txb, sizeof (txb), rxb, sizeof (rxb));

  // I2C初期化
  // クロック源の選択はライブラリで処置していないのでここで
  LPC_SYSCON->I2C1CLKSEL = SYSCON_FLEXCOMMCLKSELSRC_MAINCLK;
  IIC_Init (LPC_I2C1, IIC_Fm);
  LPC_I2C1->CFG |= I2C_CFG_MONCLKSTR;

  cls ();

  us_puts ("START\n");

  struct bme680_dev gas_sensor;
  gas_sensor.dev_id = BME680_I2C_ADDR_SECONDARY;
  gas_sensor.intf = BME680_I2C_INTF;
  gas_sensor.read =  i2c_read;
  gas_sensor.write = i2c_write;
  gas_sensor.delay_ms = Delay_ms;
  // amb_temp can be set to 25 prior to configuring the gas sensor
  // or by performing a few temperature readings without operating the gas sensor.
  gas_sensor.amb_temp = 25;
  int8_t rslt = BME680_OK;
  rslt = bme680_init (&gas_sensor);
  if (rslt == BME680_OK) us_puts ("bme680_init OK\n");

  uint8_t set_required_settings;
  // Set the temperature, pressure and humidity settings
  gas_sensor.tph_sett.os_hum = BME680_OS_2X;
  gas_sensor.tph_sett.os_pres = BME680_OS_4X;
  gas_sensor.tph_sett.os_temp = BME680_OS_8X;
  gas_sensor.tph_sett.filter = BME680_FILTER_SIZE_3;
  // Set the remaining gas sensor settings and link the heating profile
#ifdef _USE_GAS_SENSOR
  gas_sensor.gas_sett.run_gas = BME680_ENABLE_GAS_MEAS;
#else
  gas_sensor.gas_sett.run_gas = BME680_DISABLE_GAS_MEAS;
#endif
  // Create a ramp heat waveform in 3 steps
  gas_sensor.gas_sett.heatr_temp = 320; // degree Celsius
  gas_sensor.gas_sett.heatr_dur = 150; // milliseconds
  // Select the power mode
  // Must be set before writing the sensor configuration
  gas_sensor.power_mode = BME680_FORCED_MODE;
  // Set the required sensor settings needed
  set_required_settings = BME680_OST_SEL | BME680_OSP_SEL | BME680_OSH_SEL | BME680_FILTER_SEL | BME680_GAS_SENSOR_SEL;
  // Set the desired sensor configuration
  rslt = bme680_set_sensor_settings (set_required_settings,&gas_sensor);
  if (rslt == BME680_OK) us_puts ("bme680_set_sensor_settings OK\n");
  // Set the power mode
  rslt = bme680_set_sensor_mode (&gas_sensor);
  if (rslt == BME680_OK) us_puts ("bme680_set_sensor_mode OK\n");

  uint16_t meas_period;
  bme680_get_profile_dur(&meas_period, &gas_sensor);

  struct bme680_field_data data;

  if (true) {
    while (1) {
#ifdef _USE_GAS_SENSOR
      Delay_ms (2000);
#else
      Delay_ms (meas_period); // Delay till the measurement is ready
#endif
      rslt = bme680_get_sensor_data (&data, &gas_sensor);

      us_printf (
        "\rT:%1f [degC]  P:%1f [hPa]  H:%1f [%%RH]  ",
        data.temperature / 100.0f,
        data.pressure / 100.0f,
        data.humidity / 1000.0f
      );
      // Avoid using measurements from an unstable heating setup
      if(data.status & BME680_GASM_VALID_MSK)
        us_printf("G: %ld [ohms]      ", data.gas_resistance);

      // Trigger the next measurement if you would like to read data out continuously
      if (gas_sensor.power_mode == BME680_FORCED_MODE) {
        rslt = bme680_set_sensor_mode (&gas_sensor);
      }

      while (us_rx_buff ()) {
        switch (us_getc ()) {
          case '!':
            us_puts ("\n>BYE!\n ");
            Delay_ms (50);
            NVIC_SystemReset ();
            break;
        }
      }
    }
  }
}
