﻿/*
  I2C VL53L1X

  更にI2Cを推し進めてST VL53L1Xへのアクセス手段を提供するSTSW-IMG007
  なるAPIを使用して距離計測。
  しかしI2CらしいところはAPI内に記述してしまったので表に見えない。

  ・GPIO2をINPUT、GPIO3をOUTPUT、GPIO4をSCL1、GPIO5をSDA1に割り当て
  ・5V電源で使用できる秋月電子社のモジュールを試しに使用
    https://akizukidenshi.com/catalog/g/gM-14249/
    ちなみにモジュールに5V供給するとバストランスレーターにも5Vが供給
    され、DXMIOのGPIOに複数台接続するとI2Cバスがマトモに動かなくなる
    ので注意
     GPIO2 <-- GPIO
     GPIO3 --> XSHUT
     GPIO4 <-> SCL
     GPIO5 <-> SDA
  ・VL53L1Xへのアクセス手段はST社提供のSTSW-IMG007を使用
    https://www.st.com/ja/embedded-software/stsw-img007.html
  ・APIはWindowsを想定しているようなので、STSW-IMG007/API/platform 配
    下はコンパイルできる程度にI2C絡みを追記、その他整合を取った程度の
    ものを STSW-IMG007 フォルダに展開済
  ・APIをライブラリ化するためのバッチ(STSW-IMG007/API/makelib.cmd)を
    を同梱済
  ・STSW-IMG007/API/core/inc と STSW-IMG007/API/platform/inc をインク
    ルードファイル検索フォルダに、STSW-IMG007/API フォルダをライブラリ
    検索フォルダに設定すべし
  ・VL53L1Xは他のライブラリよりも前にリンクすべし
  ・IMG007は少々サイズが大きいのと複数台接続した場合の高速Ranging向き
    ではないのでSTSW-IMG009を推奨する
    https://www.st.com/ja/embedded-software/stsw-img009.html
 */

#include  <cmsis_compiler.h>
#include  <us.h>
#include  <chip.h>
// ボードの端子割り当てと最小限の初期化
#include  "e097b_def.h"

// USART_TYPEマクロで使用するUSARTと機能を選択
// CN3/CN4     1:ポーリング, 2:割り込み, 3:DMA
// CN1-6,CN1-7 4:ポーリング, 5:割り込み, 6:DMA
#ifndef USART_TYPE
  #define USART_TYPE 3
#endif
#include  "e097b_usart.h"

// I2C用関数
#include  <wl_i2c.h>
#include  <wl_delay.h>

// VL53L1用API
#include  <vl53l1_api.h>
#include  <vl53l1_platform.h>

// マクロ・グローバル変数など
//-----------------------------
// VL53L1X I2Cアドレス
#define VL53L1X_ADDRESS   (0x29)

// VL53L1XのINT信号をポーリングする場合は以下のコメントを外す
//#define VL53L1X_USEINT

//-----------------------------
// VL53L1XのGPIO1端子の状態を取得
bool GetVL53_INT (void) {
  return Chip_GPIO_ReadPortBit (LPC_GPIO_PORT, 0, _GPIO2);
}

// VL53L1XのXSHUT端子へ入力する信号を設定
static void SetVL53L1X_XSHUT (bool On) {
//  Chip_GPIO_SetPinState (LPC_GPIO_PORT, 0, _GPIO3, (bool)On);
  Chip_GPIO_SetPinState (LPC_GPIO_PORT, 0, _GPIO1, (bool)On);
}

// RangeStatusを文字列で取得
const char *GetRangeStatusString (uint8_t stat) {
  const char *st[]={
    "RANGE_VALID",
    "SIGMA_FAIL",
    "SIGNAL_FAIL",
    "RANGE_VALID_MIN_RANGE_CLIPPED",
    "OUTOFBOUNDS_FAIL",
    "HARDWARE_FAIL",
    "RANGE_VALID_NO_WRAP_CHECK_FAIL",
    "WRAP_TARGET_FAIL",
    "PROCESSING_FAIL",
    "XTALK_SIGNAL_FAIL",
    "SYNCRONISATION_INT",
    "RANGE_VALID_MERGED_PULSE",
    "TARGET_PRESENT_LACK_OF_SIGNAL",
    "MIN_RANGE_FAIL",
    "RANGE_INVALID",
    "NONE",
  };

  if (stat <= 14) return st[stat];
  return st[15];
}

//-----------------------------
void main (void) {
  const TPin pins[] = {
    { _GPIO2, PIO_TYPE_INPUT,    0,               PIO_MODE_PULLUP    }, // GPIO1
    { _GPIO3, PIO_TYPE_OUTPUT_1, 0,               PIO_MODE_DEFAULT   }, // XSHUT
    { _GPIO4, PIO_TYPE_MOVABLE,  SWM_I2C1_SCL_IO, PIO_MODE_OPENDRAIN }, // I2C1 SCL
    { _GPIO5, PIO_TYPE_MOVABLE,  SWM_I2C1_SDA_IO, PIO_MODE_OPENDRAIN }, // I2C1 SDA
  };
  // GPIO初期化
  PIO_Configure (pins, PIO_LISTSIZE (pins));

  VL53L1_Error status;
  VL53L1_Dev_t Dev;
  VL53L1_DistanceModes mode = VL53L1_DISTANCEMODE_LONG;
  VL53L1_RangingMeasurementData_t RangingData;

  uint8_t txd[100], rxd[10];
  char ch;

  // USART初期化
  us_init (Chip_Clock_GetFRGClockRate (0), 115200, USIP_8N1, txd, sizeof (txd), rxd, sizeof (rxd));
  us_printf ("Start\n");

  // I2C1クロック源の選択
  LPC_SYSCON->I2C1CLKSEL = SYSCON_FLEXCOMMCLKSELSRC_MAINCLK;

  Dev.I2Cx = LPC_I2C1;
  Dev.I2cDevAddr = VL53L1X_ADDRESS;
  Dev.comms_speed_mode = IIC_Fm;

  // VL53L1Xをハードウェアリセット
  SetVL53L1X_XSHUT (false);
  Delay_ms (2);
  SetVL53L1X_XSHUT (true);
  Delay_ms (2);

  // Ranging flowに記載がある手順
  status = VL53L1_CommsInitialise (&Dev);
  us_printf ("VL53L1_CommsInitialise = %d\n", status);
  if (status != VL53L1_ERROR_NONE) { Delay_ms (50); return; }
  status = VL53L1_WaitDeviceBooted (&Dev);
  us_printf ("VL53L1_WaitDeviceBooted = %d\n", status);
  if (status != VL53L1_ERROR_NONE) { Delay_ms (50); return; }
  status = VL53L1_DataInit (&Dev);
  us_printf ("VL53L1_DataInit = %d\n", status);
  if (status != VL53L1_ERROR_NONE) { Delay_ms (50); return; }
  status = VL53L1_StaticInit (&Dev);
  us_printf ("VL53L1_StaticInit = %d\n", status);
  if (status != VL53L1_ERROR_NONE) { Delay_ms (50); return; }
  status = VL53L1_SetDistanceMode (&Dev, mode);
  us_printf ("VL53L1_SetDistanceMode = %d\n", status);
  if (status != VL53L1_ERROR_NONE) { Delay_ms (50); return; }
  status = VL53L1_SetMeasurementTimingBudgetMicroSeconds (&Dev, 20000);  // 20~1000 *1000
  us_printf ("VL53L1_SetMeasurementTimingBudgetMicroSeconds = %d\n", status);
  if (status != VL53L1_ERROR_NONE) { Delay_ms (50); return; }
  status = VL53L1_SetInterMeasurementPeriodMilliSeconds (&Dev, 25);      // TBMS+4ms
  us_printf("VL53L1_SetInterMeasurementPeriodMilliSeconds = %d\n", status);
  if (status != VL53L1_ERROR_NONE) { Delay_ms (50); return; }
  status = VL53L1_StartMeasurement (&Dev);
  us_printf("VL53L1_StartMeasurement = %d\n", status);
  if (status != VL53L1_ERROR_NONE) { Delay_ms (50); return; }

  while (1) {
    const char *title = "Dist  SCnt  SigRmc/s  AmbRmc/s EffSpadCnt SigmaMill  Stat\n";
    us_puts (title);
    for (;;) {
      if (us_rx_buff ()) {
        switch (ch = us_getc ()) {
          case '!':
            Delay_ms (50);
            NVIC_SystemReset ();
            break;
          // 距離モードを更新
          case '1' ... '3':
            switch (ch) {
              case '1':
                mode = VL53L1_DISTANCEMODE_SHORT;
                break;
              case '2':
                mode = VL53L1_DISTANCEMODE_MEDIUM;
                break;
              case '3':
                mode = VL53L1_DISTANCEMODE_LONG;
                break;
            }
            us_printf ("\nChange mode to %d\n", mode);
            VL53L1_StopMeasurement (&Dev);
            VL53L1_SetDistanceMode (&Dev, mode);
            VL53L1_StartMeasurement (&Dev);
            us_puts (title);
        }
      }

      // 測定完了待ち
#ifdef VL53L1X_USEINT
      while (GetVL53_INT ()) Delay_ms (1);
#else
      status += VL53L1_WaitMeasurementDataReady (&Dev);
#endif
      // 測定値を取得
      status += VL53L1_GetRangingMeasurementData (&Dev, &RangingData);

      if (status != VL53L1_ERROR_NONE) {
        us_printf (
          "%4d,  %3d, %8d, %8d, %8d, %8d,  %3d(%s)                      \r",
          RangingData.RangeMilliMeter,
          RangingData.StreamCount,
          (int)(RangingData.SignalRateRtnMegaCps / 65536.0 * 1000.0),
          (int)(RangingData.AmbientRateRtnMegaCps / 65536.0 * 1000.0),
          (int)(RangingData.EffectiveSpadRtnCount /*/ 256.0 * 1000.0*/),
          (int)(RangingData.SigmaMilliMeter / 65536.0 * 100.0),

          RangingData.RangeStatus, GetRangeStatusString (RangingData.RangeStatus)
        );
      }

      // 次の測定を開始
#ifdef VL53L1X_USEINT
      status += VL53L1_ClearInterruptAndStartMeasurement (&Dev);
#else
      status += VL53L1_StartMeasurement (&Dev);
#endif
    }
    Delay_ms (20);
  }
}
